# -------------------------------------------------------------------------
# Class to hold details regarding a HTML table
# -------------------------------------------------------------------------


class TableDetails:

    def __init__(self):
        self.row_count = 0
        self.column_names = {}

    def add_column_name(self, index, name):
        """Adds a column to the table definition"""
        self.column_names[int(index)] = name

    def set_row_count(self, count):
        """Sets the row count of the table"""
        self.row_count = int(count)

    def get_column_index(self, name):
        """Gets the index position of a column with specified name"""
        for index, value in self.column_names.items():
            if value.lower() == name.lower():
                return index + 1
        raise Exception('No table column found with name: %s' % name)

    def get_row_count(self):
        """Gets the number of rows in the table"""
        return self.row_count

    def has_header_row(self):
        """Gets if table has a header row"""
        return len(self.column_names) > 0



import os
import base64
from datetime import datetime
from robot.api import logger
from accelarobotkeywords.runner import apiclient as api_clnt

HTTPPOST='POST'
HTTPGET='GET'
HTTPPUT='PUT'
HTTPPATCH='PATCH'

class TestrailIntegration: 
        
    def add_testrail_run(self, output_path, civic_platform_version, jenkins_build_number, test_rail_result_enabled, testrail_project_id, testrail_suite_id):
        is_success = None
        testrunid = None
        civp_version = os.getenv('JOB_NAME')
        if civp_version is None:
            civp_version=civic_platform_version
        enabledtestrail = str(test_rail_result_enabled)
        if enabledtestrail.lower() != 'true':
            print('TestRail Run is enabled on TestRail, No Action Taken')
            return is_success
        buildnumber = '#%s' % str(jenkins_build_number)
        testrail_file_dir = output_path
        testrail_file = os.path.join(testrail_file_dir,'testrail_run_id.txt')
        milestone_file = os.path.join(testrail_file_dir,'testrail_milestone_id.txt')
        print('Test rail test run path ' + testrail_file_dir)
        is_exits = os.path.isfile(testrail_file) and os.stat(testrail_file).st_size > 0
        if is_exits is not True:
            print('TestRail RunId file does not exists, creating new test rail run')
            now = datetime.now()
            dt_string = now.strftime("@ %m-%d-%Y %H:%M:%S")
            trail_num = dt_string
            if os.path.isfile(milestone_file) is False:
                milestone_id = self.create_testrail_milestone(civp_version, buildnumber, testrail_project_id, trail_num)
                with open(milestone_file, 'w') as writer:
                    writer.write(str(milestone_id))
            else:
                with open(milestone_file, 'r') as reader:
                    milestone_id = reader.readline()
            if milestone_id is not None:
                tc_ids = self.get_testrail_test_case_ids(civic_platform_version, testrail_project_id, testrail_suite_id)
                if tc_ids is None:
                    print('Get Test-rail IDs failed')
                    raise Exception
                if os.path.isfile(testrail_file) and os.stat(testrail_file).st_size > 0:
                    with open(testrail_file, 'r') as reader:
                        testrunid = reader.readline()
                else:
                    testrunid = self.create_testrail_testrun(civp_version, buildnumber, testrail_project_id, milestone_id, testrail_suite_id, tc_ids,trail_num)
                    with open(testrail_file, 'w') as writer:
                        writer.write(str(testrunid))
                    is_success = "TRUE"
            else:
                print('Failed to create milestone')
                raise Exception
        else:
            print('TestRail RunId file already exists, reading from file')
            with open(testrail_file, 'r') as reader:
                testrunid = reader.readline()
                print('Run Id From File: ' + str(testrunid))
    
        return testrunid

    def create_testrail_milestone(self, civp_version, deployment_environment, project_id, milestone_cnt):
        
        testrail_milestone_id = None
        test_rail_api_uri = self.get_tailrail_uri() + '/add_milestone/' + str(project_id)
        print(test_rail_api_uri)
        qparams = None
        milestone_name = civp_version + "-" + deployment_environment + "-" + str(milestone_cnt)
        body = '{"name": "' + milestone_name + '","description":"' + milestone_name + '"}'
        try:
            api_response = self.testrail_api(HTTPPOST, self.get_testrail_user(), self.get_testrail_user_password(), test_rail_api_uri, qparams, body)
            if api_response.status_code == 200:
                print(api_response)
                res_json = api_response.json()
                testrail_milestone_id = res_json['id']
                print("Milestone Id: " + str(testrail_milestone_id))
            else:
                print(api_response)
                res_json = api_response.json()
                print('Create Milestone in TestRail: Failed' + res_json)
                raise
        except Exception as e:
            print('Create Milestone in TestRail Failed')
            raise Exception(e)
        return testrail_milestone_id


    def create_testrail_testrun(self, civp_version, deployment_environment, project_id, milestone_id, test_suite_id, case_ids, testrun_cnt):
        
        testrail_testrun_id = None
        test_rail_api_uri = self.get_tailrail_uri() + '/add_run/' + str(project_id)
        print(test_rail_api_uri)
        qparams = None
        include_all = 'false'
        test_run_name = civp_version + "-" + deployment_environment + "-" + str(testrun_cnt)
        body = '{"suite_id":"' + str(test_suite_id) + '","name":"' + test_run_name + '","description":"' + test_run_name + '","milestone_id":"' + str(milestone_id) + '","include_all":' + str(include_all) + ',"case_ids":' + str(case_ids) + '}'
        try:
            api_response = self.testrail_api(HTTPPOST, self.get_testrail_user(), self.get_testrail_user_password(), test_rail_api_uri, qparams, body)
            if api_response.status_code == 200:
                print('Create TestRun in TestRail: Success')
                print(api_response)
                res_json = api_response.json()
                testrail_testrun_id = res_json['id']
                print("Milestone Id: " + str(testrail_testrun_id))
            else:
                print('Create Milestone in TestRail Failed')
                raise
        except Exception as e:
            print('Create Milestone in TestRail Failed')
            raise Exception(e)
        return testrail_testrun_id      


    def get_testrail_test_case_ids(self, civic_platform_version, project_id,test_suite_id):
        testrail_tc_ids = []
        civic_version = civic_platform_version
        test_rail_api_uri = self.get_tailrail_uri() + '/get_cases/' + str(project_id)
        print(test_rail_api_uri)
        qparams = 'suite_id=' + str(test_suite_id) + '&limit=250&offset=0'
        offsetnum=0
        body = None
        trautomationtype = 0
        if civic_version.lower().find('payment') > -1:
            trautomationtype = 1
        elif civic_version.lower().find('report') > -1:
            trautomationtype = 2
        elif civic_version.lower().find('classic') > -1:
            trautomationtype = 3
        elif civic_version.lower().find('acds') > -1:
            trautomationtype = 4
        try:
            api_response = self.testrail_api(HTTPGET, self.get_testrail_user(), self.get_testrail_user_password(), test_rail_api_uri, qparams, body)
            while api_response.status_code == 200:
                offsetnum+=250
                if len(api_response.json()['cases'])>0:
                    res_json = api_response.json()
                    for ids in res_json['cases']:
                        if civic_version.lower().find('mat') >= 0:
                            if ids['custom_robotautomated'] == 1 and ids['custom_mat'] == 1:
                                if civic_version.lower().find('aa') >= 0 and civic_version.lower().find('classic') < 0 and (ids['custom_automation_type'] == trautomationtype or ids['custom_automation_type'] == 3):
                                    testrail_tc_ids.append(ids['id'])
                                elif ids['custom_robotautomated'] == 1 and ids['custom_automation_type'] == trautomationtype:
                                    testrail_tc_ids.append(ids['id'])
                                else:   continue
                            else:   continue
                        elif civic_version.lower().find('acds') >= 0:
                            if ids['custom_robotautomated'] == 1 and ids['custom_automation_type'] == trautomationtype:
                                testrail_tc_ids.append(ids['id'])
                            else:   continue
                        elif civic_version.lower().find('aa') >= 0 and civic_version.lower().find('classic') < 0:
                            if ids['custom_robotautomated'] == 1 and (ids['custom_automation_type'] == trautomationtype or ids['custom_automation_type'] == 3):
                                testrail_tc_ids.append(ids['id'])
                            else:   continue
                        else:
                            if ids['custom_robotautomated'] == 1 and ids['custom_automation_type'] == trautomationtype:
                                testrail_tc_ids.append(ids['id'])
                            else:   continue
                else:
                    break
                qparams = 'suite_id=' + str(test_suite_id) + '&limit=250&offset=' + str(offsetnum)
                api_response = self.testrail_api(HTTPGET, self.get_testrail_user(), self.get_testrail_user_password(), test_rail_api_uri, qparams, body)
            if api_response.status_code != 200:
                print('Get test Cases from TestRail API Failed with error')
                raise
        except Exception as e:
            print('Get test Cases from TestRail failed')
            raise Exception(e)
        return testrail_tc_ids
    
    def testrail_api(self, http_method, testrail_user, testrail_user_pwd, uri, query_params, request_body):
        res_body = None
        res = None
        auth = str(
                base64.b64encode(
                    bytes('%s:%s' % (testrail_user, testrail_user_pwd), 'utf-8')), 'ascii').strip()
        headers = {'Authorization': 'Basic ' + auth}
    
        if http_method == HTTPPOST or http_method == HTTPGET:
            headers['Content-Type'] = 'application/json'
        try:
            logger.info(request_body)
            res = api_clnt.apirequest(http_method, uri, headers, query_params, request_body)
            if res.status_code == 200:
                res_body = res
            else:
                logger.error("Update Test Case  API failed " + str(request_body) + str(res))
                res_body = None
        except Exception as e:
            logger.error("Error in taro_execute_api method  : %s" % str(e))
            raise e
        return res_body
    
    def get_tailrail_uri(self):
        test_rail_uri = 'https://accela.testrail.com/index.php?/api/v2'
        return test_rail_uri


    def get_testrail_user(self):
        test_rail_user = 'mvishwanath99@accela.com'
        return test_rail_user


    def get_testrail_user_password(self):
        test_rail_user_password = 'autotest'
        return test_rail_user_password
        
        


# -----------------------------------------------------------------------
# Page/Document Conditions
#
# These are a collection of classes for WebDriverWait specific to DOM pages and documents
#
# ------------------------------------------------------------------------

from accelarobotkeywords.config.seleniumsettings import SeleniumSettings
import time

class DocumentReadyState(object):
    """An expected condition to verify the document ready state"""
    def __call__(self, driver):
        try:
            javascript = "return document.readyState"
            if driver.execute_script(javascript) == "complete":
                return True
            # above actions do not invoke the implicit wait, so sleep instead
            time.sleep(1)
            return False
        except Exception:
            return False


class PageRefresh(object):
    """An expected condition to refresh the page"""
    def __call__(self, driver):
        try:
            driver.refresh()
            return True
        except Exception:
            return False
        finally:
            driver.implicitly_wait(SeleniumSettings.get_implicit_wait())


class PageNavigateBack(object):
    """An expected condition to navigate backwards in the browser history"""
    def __call__(self, driver):
        try:
            driver.back()
            return True
        except Exception:
            return False
        finally:
            driver.implicitly_wait(SeleniumSettings.get_implicit_wait())

class PageNavigateForward(object):
    """An expected condition to navigate forwards in the browser history"""
    def __call__(self, driver):
        try:
            driver.forward()
            return True
        except Exception:
            return False
        finally:
            driver.implicitly_wait(SeleniumSettings.get_implicit_wait())



class PageGetCurrentUrl(object):
    """An expected condition to get the URL of the current page"""
    def __call__(self, driver):
        try:
            javascript = 'return window.location.href'
            url =  driver.execute_script(javascript)
            return url
        except Exception:
            return False
        finally:
            driver.implicitly_wait(SeleniumSettings.get_implicit_wait())




# -----------------------------------------------------------------------
# Element Wait Conditions
#
# These are a collection of classes for WebDriverWait specific to Elements
#
# ------------------------------------------------------------------------

import os

from selenium.common.exceptions import *
from accelarobotkeywords.config.seleniumsettings import SeleniumSettings
from selenium.webdriver.common.keys import Keys
from selenium.webdriver.common.action_chains import ActionChains


class ElementEnterText(object):
    """
    An expected condition for entering text into an element
    """

    def __init__(self, locator, text):
        self.locator_type = locator.locator_type
        self.locator_value = locator.locator_value
        self.text = text

    def __call__(self, driver):
        try:
            driver.implicitly_wait(1)
            element = driver.find_element(self.locator_type, self.locator_value)
            element.clear()
            element.send_keys(self.text)
            return True
        except Exception:
            return False
        finally:
            driver.implicitly_wait(SeleniumSettings.get_implicit_wait())


class ElementEnterTextAndCheckValueAttribute(object):
    """
    An expected condition for entering text into an element and checking value attribute
    """

    def __init__(self, locator, text):
        self.locator_type = locator.locator_type
        self.locator_value = locator.locator_value
        self.text = text

    def __call__(self, driver):
        try:
            driver.implicitly_wait(1)
            element = driver.find_element(self.locator_type, self.locator_value)
            element.clear()
            element.send_keys(self.text)
            actual_value = element.get_attribute('value')
            if actual_value is None:
                actual_value = ''
            return actual_value.strip() == self.text
        except Exception:
            return False
        finally:
            driver.implicitly_wait(SeleniumSettings.get_implicit_wait())


class ElementTypeTextIndividually(object):
    """
    An expected condition for entering text into an element by typing individual characters
    """

    def __init__(self, locator, text):
        self.locator_type = locator.locator_type
        self.locator_value = locator.locator_value
        self.text = text

    def __call__(self, driver):
        try:
            driver.implicitly_wait(1)
            element = driver.find_element(self.locator_type, self.locator_value)
            element.clear()
            for char in self.text:
                element.send_keys(char)
            actual_text = element.text
            actual_value = element.get_attribute('value')
            if actual_text is None:
                actual_text = ''
            if actual_value is None:
                actual_value = ''
            return actual_text.strip() == self.text or actual_value.strip() == self.text
        except Exception:
            return False
        finally:
            driver.implicitly_wait(SeleniumSettings.get_implicit_wait())


class ElementEnterTextWithNoClear(object):
    """
    An expected condition for entering text into an element with no clear
    """

    def __init__(self, locator, text):
        self.locator_type = locator.locator_type
        self.locator_value = locator.locator_value
        self.text = text

    def __call__(self, driver):
        try:
            driver.implicitly_wait(1)
            element = driver.find_element(self.locator_type, self.locator_value)
            element.send_keys(self.text)
            actual_text = element.text
            actual_value = element.get_attribute('value')
            if actual_text is None:
                actual_text = ''
            if actual_value is None:
                actual_value = ''
            return actual_text.strip() == self.text or actual_value.strip() == self.text
        except Exception:
            return False
        finally:
            driver.implicitly_wait(SeleniumSettings.get_implicit_wait())


class ElementEnterTextViaJavascript(object):
    """
    An expected condition for entering text into an element via javascript
    """

    def __init__(self, locator, text):
        self.locator_type = locator.locator_type
        self.locator_value = locator.locator_value
        self.text = text

    def __call__(self, driver):
        try:
            driver.implicitly_wait(1)
            element = driver.find_element(self.locator_type, self.locator_value)
            driver.execute_script('arguments[0].value=arguments[1];', element, self.text)
            return True
        except Exception:
            return False
        finally:
            driver.implicitly_wait(SeleniumSettings.get_implicit_wait())


class ElementEnterTextToAllMatching(object):
    """
    An expected condition for entering text to all matching elements
    """

    def __init__(self, locator, text):
        self.locator_type = locator.locator_type
        self.locator_value = locator.locator_value
        self.text = text

    def __call__(self, driver):
        try:
            driver.implicitly_wait(1)
            elements = driver.find_elements(self.locator_type, self.locator_value)
            for element in elements:
                element.send_keys(self.text)
            return True
        except Exception:
            return False
        finally:
            driver.implicitly_wait(SeleniumSettings.get_implicit_wait())


class ElementClearTextToAllMatching(object):
    """
    An expected condition for clearing text from all matching elements
    """

    def __init__(self, locator):
        self.locator_type = locator.locator_type
        self.locator_value = locator.locator_value

    def __call__(self, driver):
        try:
            driver.implicitly_wait(1)
            elements = driver.find_elements(self.locator_type, self.locator_value)
            for element in elements:
                element.clear()
            return True
        except Exception:
            return False
        finally:
            driver.implicitly_wait(SeleniumSettings.get_implicit_wait())


class ElementEnterTextViaActions(object):
    """
    An expected condition for entering text into an element via actions
    """

    def __init__(self, locator, text):
        self.locator_type = locator.locator_type
        self.locator_value = locator.locator_value
        self.text = text

    def __call__(self, driver):
        try:
            driver.implicitly_wait(1)
            element = driver.find_element(self.locator_type, self.locator_value)
            ActionChains(driver).click(element).send_keys(Keys.HOME).send_keys(self.text).perform()
            return True
        except Exception:
            return False
        finally:
            driver.implicitly_wait(SeleniumSettings.get_implicit_wait())


class ElementClearText(object):
    """
    An expected condition for clearing text from an element
    """

    def __init__(self, locator):
        self.locator_type = locator.locator_type
        self.locator_value = locator.locator_value

    def __call__(self, driver):
        try:
            driver.implicitly_wait(1)
            element = driver.find_element(self.locator_type, self.locator_value)
            element.clear()
            actual_text = element.text
            actual_value = element.get_attribute('value')
            if actual_text is None:
                actual_text = ''
            if actual_value is None:
                actual_value = ''
            return actual_text.strip() == '' or actual_value.strip() == ''
        except Exception:
            return False
        finally:
            driver.implicitly_wait(SeleniumSettings.get_implicit_wait())


class ElementSendAnyKey(object):
    """
    An expected condition for sending a specified special key to an element
    """

    def __init__(self, locator, key):
        self.locator_type = locator.locator_type
        self.locator_value = locator.locator_value
        self.key = key

    def __call__(self, driver):
        try:
            driver.implicitly_wait(1)
            element = driver.find_element(self.locator_type, self.locator_value)
            ActionChains(driver).click(element).send_keys(self.key).perform()
            return True
        except Exception:
            return False
        finally:
            driver.implicitly_wait(SeleniumSettings.get_implicit_wait())


class ElementSendEscapeKey(object):
    """
    An expected condition for sending Escape key to currently active element
    """

    def __call__(self, driver):
        try:
            driver.implicitly_wait(1)

            ActionChains(driver).click(driver.switch_to.active_element).send_keys(Keys.ESCAPE).perform()
            return True
        except Exception:
            return False
        finally:
            driver.implicitly_wait(SeleniumSettings.get_implicit_wait())


class ElementDoubleClick(object):
    """
    An expected condition for double clicking on an element
    """

    def __init__(self, locator):
        self.locator_type = locator.locator_type
        self.locator_value = locator.locator_value

    def __call__(self, driver):
        try:
            driver.implicitly_wait(1)
            element = driver.find_element(self.locator_type, self.locator_value)
            ActionChains(driver).move_to_element(element).double_click(element).perform()
            return True
        except Exception:
            return False
        finally:
            driver.implicitly_wait(SeleniumSettings.get_implicit_wait())


class ElementLeftClick(object):
    """
    An expected condition for left clicking on an element
    """

    def __init__(self, locator):
        self.locator_type = locator.locator_type
        self.locator_value = locator.locator_value

    def __call__(self, driver):
        try:
            driver.implicitly_wait(1)
            element = driver.find_element(self.locator_type, self.locator_value)
            element.click()
            return True
        except Exception:
            return False
        finally:
            driver.implicitly_wait(SeleniumSettings.get_implicit_wait())


class ElementWaitForDisplayed(object):
    """
    An expected condition for waiting for a displayed element
    """

    def __init__(self, locator):
        self.locator_type = locator.locator_type
        self.locator_value = locator.locator_value

    def __call__(self, driver):
        try:
            driver.implicitly_wait(1)
            element = driver.find_element(self.locator_type, self.locator_value)
            return element.is_displayed()
        except Exception:
            return False
        finally:
            driver.implicitly_wait(SeleniumSettings.get_implicit_wait())


class ElementWaitForNotDisplayed(object):
    """
    An expected condition for waiting for a not displayed element
    """

    def __init__(self, locator):
        self.locator_type = locator.locator_type
        self.locator_value = locator.locator_value

    def __call__(self, driver):
        try:
            driver.implicitly_wait(1)
            element = driver.find_element(self.locator_type, self.locator_value)
            return not element.is_displayed()
        except NoSuchElementException:
            return True
        except Exception:
            return False
        finally:
            driver.implicitly_wait(SeleniumSettings.get_implicit_wait())


class ElementWaitForEnabled(object):
    """
    An expected condition for waiting for an element to be enabled
    """

    def __init__(self, locator):
        self.locator_type = locator.locator_type
        self.locator_value = locator.locator_value

    def __call__(self, driver):
        try:
            driver.implicitly_wait(1)
            element = driver.find_element(self.locator_type, self.locator_value)
            return element.is_enabled()
        except Exception:
            return False
        finally:
            driver.implicitly_wait(SeleniumSettings.get_implicit_wait())


class ElementWaitForNotEnabled(object):
    """
    An expected condition for waiting for an element to be not enabled
    """

    def __init__(self, locator):
        self.locator_type = locator.locator_type
        self.locator_value = locator.locator_value

    def __call__(self, driver):
        try:
            driver.implicitly_wait(1)
            element = driver.find_element(self.locator_type, self.locator_value)
            return not element.is_enabled()
        except NoSuchElementException:
            return True
        except Exception:
            return False
        finally:
            driver.implicitly_wait(SeleniumSettings.get_implicit_wait())


class ElementShouldBeEnabled(object):
    """
    An expected condition for verifying that an element is enabled
    """

    def __init__(self, locator):
        self.locator_type = locator.locator_type
        self.locator_value = locator.locator_value

    def __call__(self, driver):
        try:
            element = driver.find_element(self.locator_type, self.locator_value)
            return element.is_enabled()
        except Exception:
            return False
        finally:
            driver.implicitly_wait(SeleniumSettings.get_implicit_wait())


class ElementShouldNotBeEnabled(object):
    """
    An expected condition for verifying that an element is not enabled
    """

    def __init__(self, locator):
        self.locator_type = locator.locator_type
        self.locator_value = locator.locator_value

    def __call__(self, driver):
        try:
            element = driver.find_element(self.locator_type, self.locator_value)
            return not element.is_enabled()
        except NoSuchElementException:
            return True
        except Exception:
            return False
        finally:
            driver.implicitly_wait(SeleniumSettings.get_implicit_wait())


class ElementShouldBeDisplayed(object):
    """
    An expected condition for verifying that an element is displayed
    """

    def __init__(self, locator):
        self.locator_type = locator.locator_type
        self.locator_value = locator.locator_value

    def __call__(self, driver):
        try:
            driver.implicitly_wait(1)
            element = driver.find_element(self.locator_type, self.locator_value)
            return element.is_displayed()
        except Exception:
            return False
        finally:
            driver.implicitly_wait(SeleniumSettings.get_implicit_wait())


class ElementShouldNotBeDisplayed(object):
    """
    An expected condition for verifying that an element is not displayed
    """

    def __init__(self, locator):
        self.locator_type = locator.locator_type
        self.locator_value = locator.locator_value

    def __call__(self, driver):
        try:
            driver.implicitly_wait(1)
            element = driver.find_element(self.locator_type, self.locator_value)
            return not element.is_displayed()
        except NoSuchElementException:
            return True
        except Exception:
            return False
        finally:
            driver.implicitly_wait(SeleniumSettings.get_implicit_wait())

class ElementShouldBeEmpty(object):
    """
    An expected condition for verifying that an element is empty
    """

    def __init__(self, locator):
        self.locator_type = locator.locator_type
        self.locator_value = locator.locator_value

    def __call__(self, driver):
        try:
            element = driver.find_element(self.locator_type, self.locator_value)
            actual_text = element.text
            actual_value = element.get_attribute('value')
            if actual_text is None:
                actual_text = ''
            if actual_value is None:
                actual_value = ''
            return actual_text.strip() == '' and actual_value.strip() == ''
        except Exception:
            return False
        finally:
            driver.implicitly_wait(SeleniumSettings.get_implicit_wait())


class ElementHoverAndLeftClick(object):
    """
    An expected condition to hover over and click on an element
    """

    def __init__(self, locator):
        self.locator_type = locator.locator_type
        self.locator_value = locator.locator_value

    def __call__(self, driver):
        try:
            driver.implicitly_wait(1)
            element = driver.find_element(self.locator_type, self.locator_value)
            ActionChains(driver).move_to_element(element).click().perform()
            return True
        except Exception:
            return False
        finally:
            driver.implicitly_wait(SeleniumSettings.get_implicit_wait())


class ElementGetText(object):
    """
    An expected condition to get the text of an element
    Returns a String containing the elements text
    """

    def __init__(self, locator):
        self.locator_type = locator.locator_type
        self.locator_value = locator.locator_value

    def __call__(self, driver):
        try:
            driver.implicitly_wait(1)
            element = driver.find_element(self.locator_type, self.locator_value)
            element_text = element.text
            return element_text.strip()
        except Exception:
            return False
        finally:
            driver.implicitly_wait(SeleniumSettings.get_implicit_wait())


class ElementGetTextFromAllMatchingElements(object):
    """
    An expected condition to get text from all matching elements
    Returns a List containing all text strings found for matching elements
    """

    def __init__(self, locator):
        self.locator_type = locator.locator_type
        self.locator_value = locator.locator_value

    def __call__(self, driver):
        try:
            driver.implicitly_wait(1)
            elements = driver.find_elements(self.locator_type, self.locator_value)
            text_list = []
            for element in elements:
                element_text = element.text
                text_list.append(element_text.strip())
            return text_list
        except Exception:
            return False
        finally:
            driver.implicitly_wait(SeleniumSettings.get_implicit_wait())


class ElementGetAttributeFromAllMatchingElements(object):
    """
    An expected condition to get attribute values from all matching elements
    Returns a List containing all attribute value strings found for matching elements
    """

    def __init__(self, locator, attribute_name):
        self.locator_type = locator.locator_type
        self.locator_value = locator.locator_value
        self.attribute_name = attribute_name

    def __call__(self, driver):
        try:
            driver.implicitly_wait(1)
            elements = driver.find_elements(self.locator_type, self.locator_value)
            attribute_values_list = []
            for element in elements:
                element_attribute_text = element.get_attribute(self.attribute_name)
                attribute_values_list.append(element_attribute_text.strip())
            return attribute_values_list
        except Exception:
            return False
        finally:
            driver.implicitly_wait(SeleniumSettings.get_implicit_wait())


class ElementGetTextFromChildElements(object):
    """
    An expected condition to get text from specified child element(s) of a parent locator
    Returns a List containing all text strings from child elements
    """

    def __init__(self, parent_locator, child_locator):
        self.parent_locator_type = parent_locator.locator_type
        self.parent_locator_value = parent_locator.locator_value
        self.child_locator_type = child_locator.locator_type
        self.child_locator_value = child_locator.locator_value

    def __call__(self, driver):
        try:
            driver.implicitly_wait(1)
            elements = driver.find_elements(self.parent_locator_type, self.parent_locator_value)
            text_list = []
            for element in elements:
                element_text = element.find_element(self.child_locator_type, self.child_locator_value).text
                text_list.append(element_text.strip())
            return text_list
        except Exception:
            return False
        finally:
            driver.implicitly_wait(SeleniumSettings.get_implicit_wait())


class ElementGetTextFromAllChildElements(object):
    """
    An expected condition to get text from all child element(s) of a parent locator
    Returns a List containing all text strings from child elements
    """

    def __init__(self, parent_locator):
        self.parent_locator_type = parent_locator.locator_type
        self.parent_locator_value = parent_locator.locator_value

    def __call__(self, driver):
        try:
            driver.implicitly_wait(1)
            elements = driver.find_elements(self.parent_locator_type, self.parent_locator_value)
            text_list = []
            for element in elements:
                element_text = element.text
                text_list.append(element_text.strip())
            return text_list
        except Exception:
            return False
        finally:
            driver.implicitly_wait(SeleniumSettings.get_implicit_wait())


class ElementGetAttribute(object):
    """
    An expected condition to get the attribute of an element based on its name
    Returns a String containing the element attribute's value
    """

    def __init__(self, locator, attribute_name):
        self.locator_type = locator.locator_type
        self.locator_value = locator.locator_value
        self.attribute_name = attribute_name

    def __call__(self, driver):
        try:
            driver.implicitly_wait(1)
            element = driver.find_element(self.locator_type, self.locator_value)
            element_attribute = element.get_attribute(self.attribute_name)
            return element_attribute.strip()
        except Exception:
            return False
        finally:
            driver.implicitly_wait(SeleniumSettings.get_implicit_wait())


class ElementMakeVisible(object):
    """
    An expected condition for making a hidden element visible via javascript
    """

    def __init__(self, locator):
        self.locator_type = locator.locator_type
        self.locator_value = locator.locator_value

    def __call__(self, driver):
        try:
            driver.implicitly_wait(1)
            element = driver.find_element(self.locator_type, self.locator_value)
            driver.execute_script('arguments[0].style = ""; arguments[0].style.display = "block"; arguments['
                                  '0].style.visibility = "visible";', element)
            return True
        except Exception:
            return False
        finally:
            driver.implicitly_wait(SeleniumSettings.get_implicit_wait())


class ElementMakeHidden(object):
    """
    An expected condition for making an element hidden via javascript
    """

    def __init__(self, locator):
        self.locator_type = locator.locator_type
        self.locator_value = locator.locator_value

    def __call__(self, driver):
        try:
            driver.implicitly_wait(1)
            element = driver.find_element(self.locator_type, self.locator_value)
            driver.execute_script('arguments[0].style = ""; arguments[0].style.display = "none"; arguments['
                                  '0].style.visibility = "hidden";', element)
            return True
        except Exception:
            return False
        finally:
            driver.implicitly_wait(SeleniumSettings.get_implicit_wait())


class ElementClickNthElement(object):
    """
    An expected condition Click on nth position element,
    if a locator has multiple matching element, it will click on the element based on the
    index(nth position ) passed in
    """

    def __init__(self, locator, nth_position):
        self.locator_type = locator.locator_type
        self.locator_value = locator.locator_value
        self.nth_position = int(nth_position)

    def __call__(self, driver):
        try:
            driver.implicitly_wait(1)
            elements = driver.find_elements(self.locator_type, self.locator_value)
            elements[self.nth_position].click()
            return True
        except Exception:
            return False
        finally:
            driver.implicitly_wait(SeleniumSettings.get_implicit_wait())


class ElementControlAndLeftClick(object):
    """
    An expected condition to Control and left click on an element
    ctrl + Left click
    """

    def __init__(self, locator):
        self.locator_type = locator.locator_type
        self.locator_value = locator.locator_value

    def __call__(self, driver):
        try:
            driver.implicitly_wait(1)
            element = driver.find_element(self.locator_type, self.locator_value)
            ActionChains(driver).key_down(Keys.CONTROL).click(element).key_up(Keys.CONTROL).perform()
            return True
        except Exception:
            return False
        finally:
            driver.implicitly_wait(SeleniumSettings.get_implicit_wait())


class ElementControlAndRightClick(object):
    """
    An expected condition to Control and Right click on an element
    ctrl + Right click
    """

    def __init__(self, locator):
        self.locator_type = locator.locator_type
        self.locator_value = locator.locator_value

    def __call__(self, driver):
        try:
            driver.implicitly_wait(1)
            element = driver.find_element(self.locator_type, self.locator_value)
            ActionChains(driver).key_down(Keys.CONTROL).context_click(element).key_up(Keys.CONTROL).perform()
            return True
        except Exception:
            return False
        finally:
            driver.implicitly_wait(SeleniumSettings.get_implicit_wait())


class ElementControlAndPressKey(object):
    """
    An expected condition to Control and Press a key on the in focus element
    ctrl + press a,v,x,b key
    """

    def __init__(self, key):
        self.key = key

    def __call__(self, driver):
        try:
            driver.implicitly_wait(1)
            ActionChains(driver).key_down(Keys.CONTROL).send_keys(self.key).key_up(Keys.CONTROL).perform()
            return True
        except Exception:
            return False
        finally:
            driver.implicitly_wait(SeleniumSettings.get_implicit_wait())


class ElementAltAndPressKey(object):
    """
    An expected condition to Alt and Press a key on the in focus element
    Alt + press key
    """

    def __init__(self, key):
        self.key = key

    def __call__(self, driver):
        try:
            driver.implicitly_wait(1)
            ActionChains(driver).key_down(Keys.ALT).send_keys(self.key).key_up(Keys.ALT).perform()
            return True
        except Exception:
            return False
        finally:
            driver.implicitly_wait(SeleniumSettings.get_implicit_wait())


class ElementJsClick(object):
    """
    An expected condition to JS click (javascript click ) on an element
    """

    def __init__(self, locator):
        self.locator_type = locator.locator_type
        self.locator_value = locator.locator_value

    def __call__(self, driver):
        try:
            driver.implicitly_wait(1)
            element = driver.find_element(self.locator_type, self.locator_value)
            driver.execute_script("arguments[0].click();", element)
            return True
        except Exception:
            return False
        finally:
            driver.implicitly_wait(SeleniumSettings.get_implicit_wait())


class ElementGetTotalNumberOfElements(object):
    """
    An expected condition to get total number of elements on a locator
    """
    def __init__(self, locator):
        self.locator_type = locator.locator_type
        self.locator_value = locator.locator_value

    def __call__(self, driver):
        try:
            driver.implicitly_wait(1)
            elements = driver.find_elements(self.locator_type, self.locator_value)
            return str(len(elements))
        except Exception:
            return False
        finally:
            driver.implicitly_wait(SeleniumSettings.get_implicit_wait())


class ElementSetFocus(object):
    """
    An expected condition to set focus on an element
    """
    def __init__(self, locator):
        self.locator_type = locator.locator_type
        self.locator_value = locator.locator_value

    def __call__(self, driver):
        try:
            driver.implicitly_wait(1)
            element = driver.find_element(self.locator_type, self.locator_value)
            element.send_keys(Keys.NULL)
            return True
        except Exception:
            return False
        finally:
            driver.implicitly_wait(SeleniumSettings.get_implicit_wait())


class ElementGetActiveElement(object):
    """
    An expected condition to get currently active/focused element
    """

    def __call__(self, driver):
        try:
            driver.implicitly_wait(1)
            elem = driver.switch_to.active_element
            return elem
        except Exception:
            return False
        finally:
            driver.implicitly_wait(SeleniumSettings.get_implicit_wait())


class ElementSetAttributeValueByJS(object):
    """
    An expected condition to set attribute value by js: the attribute and value both are passed in
    """
    def __init__(self, locator, attribute, value):
        self.locator_type = locator.locator_type
        self.locator_value = locator.locator_value
        self.attribute = attribute
        self.value = value

    def __call__(self, driver):
        try:
            driver.implicitly_wait(1)
            element = driver.find_element(self.locator_type, self.locator_value)
            script = ("arguments[0].%s=arguments[1];" % self.attribute)
            driver.execute_script(script, element, self.value)
            return True
        except Exception:
            return False
        finally:
            driver.implicitly_wait(SeleniumSettings.get_implicit_wait())


class ElementMouseOver(object):
    """
    An expected condition to move mouse over to the element
    """

    def __init__(self, locator):
        self.locator_type = locator.locator_type
        self.locator_value = locator.locator_value

    def __call__(self, driver):
        try:
            driver.implicitly_wait(1)
            element = driver.find_element(self.locator_type, self.locator_value)
            ActionChains(driver).move_to_element(element).perform()
            return True
        except Exception:
            return False
        finally:
            driver.implicitly_wait(SeleniumSettings.get_implicit_wait())


class ElementGetCSSPropertyValue(object):
    """
    An expected condition to Get value of css property. eg. font-size
    """
    def __init__(self, locator, css_property):
        self.locator_type = locator.locator_type
        self.locator_value = locator.locator_value
        self.css_property = css_property

    def __call__(self, driver):
        try:
            driver.implicitly_wait(1)
            css_value = driver.find_element(self.locator_type, self.locator_value).value_of_css_property(self.css_property)
            return str(css_value)
        except Exception:
            return False
        finally:
            driver.implicitly_wait(SeleniumSettings.get_implicit_wait())


class ElementDragAndDrop(object):
    """
    An expected condition drag and drop element to another element
    """
    def __init__(self, locator_from, locator_to):
        self.locator_from_type = locator_from.locator_type
        self.locator_from_value = locator_from.locator_value
        self.locator_to_type = locator_to.locator_type
        self.locator_to_value = locator_to.locator_value

    def __call__(self, driver):
        try:
            driver.implicitly_wait(1)
            element_from = driver.find_element(self.locator_from_type, self.locator_from_value)
            element_to = driver.find_element(self.locator_to_type, self.locator_to_value)
            ActionChains(driver).click_and_hold(element_from).perform()
            ActionChains(driver).move_to_element(element_to).perform()
            ActionChains(driver).release(element_to).perform()
            return True
        except Exception:
            return False
        finally:
            driver.implicitly_wait(SeleniumSettings.get_implicit_wait())


class ElementDragAndDropXYPosition(object):
    """
    An expected condition drag and drop an element to x,y offset
    """
    def __init__(self, locator_from, x_offset, y_offset):
        self.locator_from_type = locator_from.locator_type
        self.locator_from_value = locator_from.locator_value
        self.x_offset = int(x_offset)
        self.y_offset = int(y_offset)

    def __call__(self, driver):
        try:
            driver.implicitly_wait(1)
            element_from = driver.find_element(self.locator_from_type, self.locator_from_value)
            ActionChains(driver).drag_and_drop_by_offset(element_from, self.x_offset, self.y_offset)
            return True
        except Exception:
            return False
        finally:
            driver.implicitly_wait(SeleniumSettings.get_implicit_wait())


class ElementChooseFile(object):
    """
    An expected condition for entering text into a choose file input element
    """

    def __init__(self, locator, file_path):
        self.locator_type = locator.locator_type
        self.locator_value = locator.locator_value
        self.text = file_path

    def __call__(self, driver):
        try:
            driver.implicitly_wait(1)
            self.info('Sending %s to browser.' % os.path.abspath(self.text))
            element = driver.find_element(self.locator_type, self.locator_value)
            element.clear()
            element.send_keys(self.text)
            return True
        except Exception:
            return False
        finally:
            driver.implicitly_wait(SeleniumSettings.get_implicit_wait())


class ScrollToElementViaJavascript(object):
    """
    An expected condition for scrolling to an element via javascript
    """

    def __init__(self, locator):
        self.locator_type = locator.locator_type
        self.locator_value = locator.locator_value

    def __call__(self, driver):
        try:
            driver.implicitly_wait(1)
            element = driver.find_element(self.locator_type, self.locator_value)
            driver.execute_script('arguments[0].scrollIntoView(true);', element)
            return True
        except Exception:
            return False
        finally:
            driver.implicitly_wait(SeleniumSettings.get_implicit_wait())

class JsScriptExecutor(object):
    """
    For a general script to be executed
    """
    def __init__(self, js_script):
        self.js_script = js_script

    def __call__(self, driver):
        try:
            driver.implicitly_wait(1)
            driver.execute_script(self.js_script)
            return True
        except Exception:
            return False
        finally:
            driver.implicitly_wait(SeleniumSettings.get_implicit_wait())

class ElementGetCssValue(object):
    """
    An expected condition to get the css value of an element
    Returns a String containing the elements css value
    """

    def __init__(self, locator, css_value):
        self.locator_type = locator.locator_type
        self.locator_value = locator.locator_value
        self.css_value = css_value

    def __call__(self, driver):
        try:
            driver.implicitly_wait(1)
            element = driver.find_element(self.locator_type, self.locator_value)
            element_text = element.getCssValue(self.css_value)
            return element_text.strip()
        except Exception:
            return False
        finally:
            driver.implicitly_wait(SeleniumSettings.get_implicit_wait())


# ---------------------------------------------------------------------------------
# Table Wait Conditions
#
# These are a collection of classes for WebDriverWait specific to Tables
#
# --------------------------------------------------------------------------------

from accelarobotkeywords.config.seleniumsettings import SeleniumSettings
from accelarobotkeywords.framework.selenium.table.tablescanner import TableScanner


class GetTableDetails(object):
    """Condition to get the tables details"""
    def __init__(self, locator):
        self.locator = locator

    def __call__(self, driver):
        try:
            driver.implicitly_wait(1)
            table_info = TableScanner.get_table_details(driver, self.locator)
            return table_info
        except Exception:
            return False
        finally:
            driver.implicitly_wait(SeleniumSettings.get_implicit_wait())


class GetCellText(object):
    """Condition to get the text from a cell"""
    def __init__(self, locator, row_number, column_number):
        self.locator_type = locator.locator_type
        self.locator_value = locator.locator_value
        self.row_num = row_number
        self.col_num = column_number

    def __call__(self, driver):
        try:
            driver.implicitly_wait(1)
            table_element = driver.find_element(self.locator_type, self.locator_value)
            xpath = 'descendant::tr[%s]' % self.row_num + '/descendant::td[%s]' % self.col_num
            cell_element = table_element.find_element(by='xpath', value=xpath)
            return cell_element.text
        except Exception:
            return False
        finally:
            driver.implicitly_wait(SeleniumSettings.get_implicit_wait())


class SearchForCellValue(object):
    """Condition to search text value in a column"""
    def __init__(self, locator, column_index, search_text, start_row, max_rows):
        self.locator = locator
        self.column_index = column_index
        self.search_text = search_text
        self.max_rows = max_rows
        self.start_row = start_row

    def __call__(self, driver):
        try:
            driver.implicitly_wait(1)
            row_index = TableScanner.search_column_for_value(driver=driver, locator=self.locator,
                                                             column_index=self.column_index,
                                                             search_text=self.search_text, max_rows=self.max_rows,
                                                             start_row=self.start_row)
            return row_index
        except Exception:
            return False
        finally:
            driver.implicitly_wait(SeleniumSettings.get_implicit_wait())


class ClickOnCell(object):
    """Condition to search text value in column and click on child element in cell"""
    def __init__(self, locator, column_index, search_text, start_row, max_rows, click_column_index):
        self.locator = locator
        self.column_index = column_index
        self.search_text = search_text
        self.max_rows = max_rows
        self.start_row = start_row
        self.click_column_index = click_column_index

    def __call__(self, driver):
        try:
            driver.implicitly_wait(1)
            row_index = TableScanner.search_column_for_value(driver=driver, locator=self.locator,
                                                             column_index=self.column_index,
                                                             search_text=self.search_text, max_rows=self.max_rows,
                                                             start_row=self.start_row)
            table_element = driver.find_element(self.locator.locator_type, self.locator.locator_value)
            xpath = 'descendant::tr[%s]' % row_index + '/descendant::td[%s]' % self.click_column_index + '/*[1]'
            element = table_element.find_element(by='xpath', value=xpath)
            element.click()
            return True
        except Exception:
            return False
        finally:
            driver.implicitly_wait(SeleniumSettings.get_implicit_wait())



# -----------------------------------------------------------------------
# Window Wait Conditions
#
# These are a collection of classes for WebDriverWait specific to Windows
#
# ------------------------------------------------------------------------

from accelarobotkeywords.config.seleniumsettings import SeleniumSettings
import time
from robot.api import logger


class WindowSwitchToWindowWithTitle(object):
    """An expected condition to wait for a window with a specified title then switch to it"""
    def __init__(self, title):
        self.title = title

    def __call__(self, driver):
        try:
            driver.implicitly_wait(1)
            handles = driver.window_handles
            for handle in handles:
                driver.switch_to.window(handle)
                window_title = driver.title
                if window_title.casefold() == self.title.casefold():
                    return True
            # above actions do not invoke the implicit wait, so sleep instead
            time.sleep(1)
            return False
        except Exception:
            return False
        finally:
            driver.implicitly_wait(SeleniumSettings.get_implicit_wait())


class WindowSwitchToWindowWithUrl(object):
    """An expected condition to wait for a window with a specified url then switch to it"""
    def __init__(self, url):
        self.url = url

    def __call__(self, driver):
        try:
            driver.implicitly_wait(1)
            handles = driver.window_handles
            for handle in handles:
                driver.switch_to.window(handle)
                window_url = driver.current_url
                if self.url in window_url:
                    return True
            # above actions do not invoke the implicit wait, so sleep instead
            time.sleep(1)
            return False
        except Exception:
            return False
        finally:
            driver.implicitly_wait(SeleniumSettings.get_implicit_wait())


class WindowSwitchToWindowWithHandleIndex(object):
    """switches to window based on the handle index passed on"""
    def __init__(self, handle_index):
        self.index = handle_index
        logger.info(handle_index)

    def __call__(self, driver):
        try:
            driver.implicitly_wait(1)
            handles = driver.window_handles
            if len(handles) > 1:
                driver.switch_to.window(handles[int(self.index)])
                return True
            # above actions do not invoke the implicit wait, so sleep instead
            time.sleep(1)
            return False
        except Exception:
            return False
        finally:
            driver.implicitly_wait(SeleniumSettings.get_implicit_wait())


class WindowSwitchToFirstFoundWindow(object):
    """An expected condition to wait for a second window and switch to it"""
    def __init__(self, default_handle):
        self.default_handle = default_handle

    def __call__(self, driver):
        try:
            driver.implicitly_wait(1)
            handles = driver.window_handles
            for handle in handles:
                if handle != self.default_handle:
                    driver.switch_to.window(handle)
                    return True
            # above actions dont invoke an implicit wait
            time.sleep(1)
            return False
        except Exception:
            return False
        finally:
            driver.implicitly_wait(SeleniumSettings.get_implicit_wait())


class WindowMaximizeWindow(object):
    """Maximize the window"""
    def __call__(self, driver):
        try:
            driver.maximize_window()
            time.sleep(1)
            return True
        except Exception:
            return False
        finally:
            driver.implicitly_wait(SeleniumSettings.get_implicit_wait())


class WindowTitleShouldMatch(object):
    """An expected condition to match window title with passed in title, title can be partially match """
    def __init__(self, title):
        self.title = title

    def __call__(self, driver):
        try:
            driver.implicitly_wait(1)
            window_title = driver.title
            if window_title.casefold() in self.title.casefold() or self.title.casefold() in window_title.casefold():
                return True
            # above actions do not invoke the implicit wait, so sleep instead
            time.sleep(1)
            return False
        except Exception:
            return False
        finally:
            driver.implicitly_wait(SeleniumSettings.get_implicit_wait())


class WindowTitleShouldBeEqual(object):
    """An expected condition to match window title with passed in title, title has to be exact match """
    def __init__(self, title):
        self.title = title

    def __call__(self, driver):
        try:
            driver.implicitly_wait(1)
            window_title = driver.title
            if window_title.casefold() == self.title.casefold():
                return True
            # above actions do not invoke the implicit wait, so sleep instead
            time.sleep(1)
            return False
        except Exception:
            return False
        finally:
            driver.implicitly_wait(SeleniumSettings.get_implicit_wait())


class WindowCloseWindowBasedOnTitle(object):
    """An expected condition to close window based on title, title can be partially match """
    def __init__(self, title):
        self.title = title

    def __call__(self, driver):
        try:
            driver.implicitly_wait(1)
            window_title = driver.title
            if window_title.casefold() in self.title.casefold() or self.title.casefold() in window_title.casefold():
                driver.close()
                return True
            # above actions do not invoke the implicit wait, so sleep instead
            time.sleep(1)
            return False
        except Exception:
            return False
        finally:
            driver.implicitly_wait(SeleniumSettings.get_implicit_wait())


class WindowCloseWindowBasedOnTitleEqual(object):
    """An expected condition to close window based on title, title has to be exact match """
    def __init__(self, title):
        self.title = title

    def __call__(self, driver):
        try:
            driver.implicitly_wait(1)
            window_title = driver.title
            if window_title.casefold() == self.title.casefold():
                driver.close()
                return True
            # above actions do not invoke the implicit wait, so sleep instead
            time.sleep(1)
            return False
        except Exception:
            return False
        finally:
            driver.implicitly_wait(SeleniumSettings.get_implicit_wait())


class WindowCloseWindowBasedOnUrl(object):
    """An expected condition to close window based on url, url can be partially match """
    def __init__(self, url):
        self.url = url

    def __call__(self, driver):
        try:
            driver.implicitly_wait(1)
            window_url = driver.current_url
            if window_url in self.url or self.url in window_url:
                driver.close()
                return True
            # above actions do not invoke the implicit wait, so sleep instead
            time.sleep(1)
            return False
        except Exception:
            return False
        finally:
            driver.implicitly_wait(SeleniumSettings.get_implicit_wait())


class WindowSwitchToMainWindow(object):
    """An expected condition to switch to main/base window, window that is opened first at the beginning of test"""
    def __init__(self, default_handle):
        self.default_handle = default_handle

    def __call__(self, driver):
        try:
            driver.implicitly_wait(1)
            handles = driver.window_handles
            for handle in handles:
                if handle == self.default_handle:
                    driver.switch_to.window(handle)
                    return True
            # above actions dont invoke an implicit wait
            time.sleep(1)
            return False
        except Exception:
            return False
        finally:
            driver.implicitly_wait(SeleniumSettings.get_implicit_wait())

class WindowSwitchToNewWindow(object):
    """
    An expected condition to open a new window or tab.
    """
    def __init__(self, window_type):
        self.window_type = window_type

    def __call__(self, driver):
        try:
            driver.implicitly_wait(1)
            # switch to a new window or tab
            driver.switch_to.new_window(self.window_type)
            return True
        except Exception:
            return False
        finally:
            driver.implicitly_wait(SeleniumSettings.get_implicit_wait())
            
class WindowGetWindowTitle(object):
    """Gets the title of a window - returns the title text in a string"""
    def __call__(self, driver):
        try:
            driver.implicitly_wait(1)
            window_title = driver.title
            return window_title.strip()
        except Exception:
            return False
        finally:
            driver.implicitly_wait(SeleniumSettings.get_implicit_wait())

class WindowGetWindowHandles(object):
    """Gets the number of windows that are open - returns the total count"""
    def __call__(self, driver):
        try:
            driver.implicitly_wait(1)
            handles = driver.window_handles
            total_windows = len(handles)
            return int(total_windows)
        except Exception:
            return False
        finally:
            driver.implicitly_wait(SeleniumSettings.get_implicit_wait())



# ------------------------------------------------------------------------------
# Class to load settings from the displaysettings.json file
# ------------------------------------------------------------------------------

import json
import os
from robot.api import logger


class DisplaySettings(object):
    """
    Class that loads and exposes the display settings from the displaysettings.json file
    """
    __json_data = None
    __width = None
    __height = None
    __depth = None

    @staticmethod
    def __get_json_value(key):
        if DisplaySettings.__json_data is None:
            filename = os.path.abspath(os.path.join(os.path.dirname(__file__), 'displaysettings.json'))
            with open(filename, 'r') as json_file:
                raw_data = json_file.read()
                DisplaySettings.__json_data = json.loads(raw_data)
                logger.info("Display settings json file loaded")
        return DisplaySettings.__json_data[key]

    @staticmethod
    def get_width():
        if DisplaySettings.__width is None:
            result = DisplaySettings.__get_json_value('width')
            DisplaySettings.__width = int(result)
        logger.info("display width = %s" % str(DisplaySettings.__width))
        return DisplaySettings.__width

    @staticmethod
    def get_height():
        if DisplaySettings.__height is None:
            result = DisplaySettings.__get_json_value('height')
            DisplaySettings.__height = int(result)
        logger.info("display height = %s" % str(DisplaySettings.__height))
        return DisplaySettings.__height

    @staticmethod
    def get_depth():
        if DisplaySettings.__depth is None:
            result = DisplaySettings.__get_json_value('depth')
            DisplaySettings.__depth = int(result)
        logger.info("display depth = %s" % str(DisplaySettings.__depth))
        return DisplaySettings.__depth

    @staticmethod
    def set_width(width):
        DisplaySettings.__width = width

    @staticmethod
    def set_height(height):
        DisplaySettings.__height = height

    @staticmethod
    def set_depth(depth):
        DisplaySettings.__depth = depth



# ----------------------------------------------------------------------------------------
# DriverFactory
#
# Class for creating a browser instance
# The options are:
# - Local browser (on windows)
# - Local browser (on Linux)
# - Local browser (on Mac)
#
# OS detection is automatic
#
# Chrome driver locations:
# - Linux, assume the chrome driver is already installed as part of docker container
# - Mac, assume the chrome driver is already part of the path
# - Windows, chrome driver exe is part of the python package
#
# Chrome downloads directory:
# - This is set to $ROBOTOUT/downloads
# -----------------------------------------------------------------------------------------

from robot.api import logger
from accelarobotkeywords.config.seleniumsettings import SeleniumSettings
from accelarobotkeywords.config.operatingsystems import OperatingSystem
from selenium import webdriver
from os import path
from accelarobotkeywords.framework.framework import Framework
from accelarobotkeywords.config.displaysettings import DisplaySettings
from accelarobotkeywords.config.browserlocations import BrowserLocations
from accelarobotkeywords.config.desiredcapabilities import DesiredCapabilities
from robot.libraries.BuiltIn import BuiltIn
from selenium.webdriver.chrome.service import Service
# from selenium.webdriver.common.by import By
import os

class DriverFactory:

    @staticmethod
    def create_browser():
        """Creates a chrome instance and returns the driver object"""
        logger.debug("Checking the value of IsLamdaTest tag")
        remote_browser = "No"
        global variables
        variables = BuiltIn().get_variables()
        try:
            testTags = variables['@{TEST_TAGS}']
            for s in range(len(testTags)):
                if testTags[s].split(":")[0] == "IsLamdaTest":
                    print("Value of IsLamdaTest is: {}".format(testTags[s].split(":")[1]))
                    remote_browser = testTags[s].split(":")[1]
        except:
            logger.info("TEST_TAGS are not set")
        try:
            browser_locations = os.environ.get('BROWSER_LOCATION')
            logger.debug('BROWSER_LOCATION is ' + str(browser_locations))
            if BrowserLocations.get_browser_location(browser_locations) == BrowserLocations.Remote:
                desired_capabilities = os.environ.get('BROWSER_CAPABILITIES')
                job_name = os.environ.get('REPORTTITLE')
                build_number = os.environ.get('JENKINSBUILDNUMBER')
                logger.debug('BROWSER_CAPABILITIES is ' + str(desired_capabilities))
                if DesiredCapabilities.get_browser_desired_capabilities(desired_capabilities) is not None:
                    customcapabilities = DesiredCapabilities.get_browser_desired_capabilities(desired_capabilities)
                    if job_name is None or build_number is None:
                        customcapabilities['build'] = variables['${TEST_NAME}']
                    else:
                        customcapabilities['build'] = "%s%s" % (job_name,build_number)
                    customcapabilities['name'] = variables['${TEST_NAME}']
                    driver = webdriver.Remote(
                            desired_capabilities=customcapabilities,
                            command_executor=DriverFactory.__get_grid_url()
                        )
                else:
                    raise Exception('Browser capabilitise not defined')
            else:
                logger.debug("Creating chrome instance")
                system = OperatingSystem.get_operating_system()
                if system == OperatingSystem.Windows:
                    driver = DriverFactory.__create_local_windows()
                elif system == OperatingSystem.Linux:
                    driver = DriverFactory.__create_local_linux()
                elif system == OperatingSystem.Mac:
                    driver = DriverFactory.__create_local_mac()
                else:
                    raise Exception("Unsupported browser location and operating system")
                browser_name = SeleniumSettings.get_browser_name().lower()
                network_logger_enabled = os.environ.get('NETWORK_LOGGER_ENABLED', '').upper() == 'TRUE'
                if network_logger_enabled and browser_name in ('chrome', 'edge'):
                    try:
                        driver.execute_cdp_cmd("Network.enable", {})
                        logger.info("CDP Network domain enabled for NetworkLogger")
                    except Exception as e:
                        logger.warning("Failed to enable CDP Network domain: %r", e)
                driver.implicitly_wait(SeleniumSettings.get_implicit_wait())
                driver.set_script_timeout(SeleniumSettings.get_script_wait())
                driver.set_page_load_timeout(SeleniumSettings.get_page_load_wait())

            return driver
        except Exception as e:
            logger.error("Error creating chrome instance: %s" % e)
            raise

    @staticmethod
    def __create_local_windows():
        """Creates chrome instance on local windows"""
        try:
            if (SeleniumSettings.get_browser_name() == 'edge'):
                options = DriverFactory.__create_edge_options()
                driver_path = DriverFactory.__get_driver_location(OperatingSystem.Windows)
                driver = webdriver.Edge(service=Service(driver_path), options=options)
                logger.info("Edge driver instance created for Windows")
            elif (SeleniumSettings.get_browser_name() == 'firefox'):
                driver_path = DriverFactory.__get_driver_location(OperatingSystem.Windows)
                driver = webdriver.Firefox(service=Service(driver_path))
                logger.info("Firefox driver instance created for Windows")
            else:
                options = DriverFactory.__create_options()
                driver_path = DriverFactory.__get_driver_location(OperatingSystem.Windows)
                driver = webdriver.Chrome(service=Service(driver_path), options=options)
                logger.info("Chrome driver instance created for Windows")
            return driver
        except Exception as e:
            logger.error("Unable to create chrome driver on Windows: %s" % e)
            raise

    @staticmethod
    def __create_local_linux():
        """Creates chrome instance on local linux"""
        try:
            options = DriverFactory.__create_options()
            driver_path = DriverFactory.__get_driver_location(OperatingSystem.Linux)
            driver = webdriver.Chrome(service=Service(driver_path), options=options)
            logger.info("Chrome driver instance created for linux")
            return driver
        except Exception as e:
            logger.error("Unable to create chrome driver on Linux: %s" % e)
            raise

    @staticmethod
    def __create_local_mac():
        """Creates chrome instance on local mac"""
        try:
            if (SeleniumSettings.get_browser_name() == 'edge'):
                options = DriverFactory.__create_edge_options()
                driver_path = DriverFactory.__get_driver_location(OperatingSystem.Mac)
                driver = webdriver.Edge(service=Service(driver_path), options=options)
                logger.info("Edge driver instance created for mac")
            else:
                options = DriverFactory.__create_options()
                driver_path = DriverFactory.__get_driver_location(OperatingSystem.Mac)
                driver = webdriver.Chrome(service=Service(driver_path), options=options)
                logger.info("Chrome driver instance created for Mac")
            return driver
        except Exception as e:
            logger.error("Unable to create chrome driver on Mac: %s" % e)
            raise

    @staticmethod
    def __create_options():
        """Sets up the chrome options"""
        options = webdriver.ChromeOptions()
        downloads = Framework.get_downloads_path()
        options.add_argument("--allow-running-insecure-content")
        options.add_argument("--disable-plugins")
        options.add_argument("--disable-extensions")
        options.add_argument("--disable-infobars")
        options.add_argument("--start-maximized")
        options.add_argument("--no-sandbox")
        options.add_argument("--disable-background-networking")
        preferences = {
            "download.default_directory": downloads,
            "download.prompt_for_download": False,
            "profile.default_content_setting_values.automatic_downloads": 1,
            "credentials_enable_service": False,
            "profile.password_manager_enabled": False
        }
        system = OperatingSystem.get_operating_system()
        if system == OperatingSystem.Linux:
            preferences["download.prompt_for_download"] = False
            preferences["download.directory_upgrade"] = True
            width = DisplaySettings.get_width()
            height = DisplaySettings.get_height()
            options.add_argument("--window-size=%s,%s" % (width, height))
        options.add_experimental_option("prefs", preferences)
        options.add_experimental_option('useAutomationExtension', False)
        options.add_experimental_option('excludeSwitches', ['enable-automation'])
        network_logger_enabled = os.environ.get('NETWORK_LOGGER_ENABLED', '').upper() == 'TRUE'
        if network_logger_enabled:
            options.set_capability('goog:loggingPrefs', {'performance': 'ALL'})
            logger.info("Performance logging enabled")
        else:
            logger.info("Performance logging disabled")
        zap_proxy_enabled = os.environ.get('ZAP_ENABLED')
        if zap_proxy_enabled != None:
            if zap_proxy_enabled.upper() == "TRUE":
                zap_proxy_host = os.environ.get('PROXY_HOST')
                zap_proxy_port = os.environ.get('PROXY_PORT')
                zap_proxy_host_port = "http://" + zap_proxy_host + ":" + str(zap_proxy_port)
                zap_proxy_host_port_https = "https://" + zap_proxy_host + ":" + str(zap_proxy_port)
                options.add_argument('--proxy-server=%s' % zap_proxy_host_port)
                options.add_argument('ignore-certificate-errors')
        return options
    
    @staticmethod
    def __create_edge_options():
        """Sets up the edge options"""
        options = webdriver.EdgeOptions()
        options.add_argument("--start-maximized") 
        options.add_argument("--disable-extensions")
        options.add_argument("--disable-popup-blocking") 
        options.add_experimental_option("excludeSwitches", ["enable-automation"])
        options.add_experimental_option("useAutomationExtension", False)
        prefs = {
            "download.default_directory": os.path.join(os.path.expanduser("~"), "Downloads"),
            "profile.default_content_settings.popups": 0,
            "credentials_enable_service": False,
            "profile.password_manager_enabled": False
        }
        options.add_experimental_option("prefs", prefs)
        return options

    @staticmethod
    def __get_driver_location(system):
        """Gets the location of the driver executable"""
        if system == OperatingSystem.Linux:
            driver_path = '/usr/bin/chromedriver'
            logger.info('Driver path: %s' % driver_path)
            return driver_path
        elif system == OperatingSystem.Mac:
            drivers_dir_path = path.join(path.dirname(path.realpath(__file__)), 'drivers')
            if SeleniumSettings.get_browser_name() == 'edge':
                driver_path = path.join(drivers_dir_path, 'msedgedriver')
            else:
                driver_path = path.join(drivers_dir_path, 'chromedriver')
            logger.info('Driver path: %s' % driver_path)
            return driver_path
        elif system == OperatingSystem.Windows:
            drivers_dir_path = path.join(path.dirname(path.realpath(__file__)), 'drivers')
            if SeleniumSettings.get_browser_name() == 'edge':
                driver_path = path.join(drivers_dir_path, 'msedgedriver.exe')
            elif SeleniumSettings.get_browser_name() == 'firefox':
                driver_path = path.join(drivers_dir_path, 'geckodriver.exe')
            else:
                driver_path = path.join(drivers_dir_path, 'chromedriver.exe')
            logger.info('Driver path: %s' % driver_path)
            return driver_path
        else:
            raise Exception("Driver path cannot be created for current OS")

    @staticmethod
    def __get_capabilities():
        """Returns the capabilities for the remote driver"""
        desired_cap = {
            "build": variables['${SUITE NAME}'],
            "name": variables['${TEST_NAME}'],
            "platformName": "Windows 10",
            "browserName": "Chrome",
            "version": "91.0",
            "resolution": "1440x900"
        }
        return desired_cap

    @staticmethod
    def __get_grid_url():
        """Gets Grid url of the lamda test server"""
        # username: Username can be found at automation dashboard
        username = "mvishwanath"
        # accessToken:  AccessToken can be generated from automation dashboard or profile section
        accessToken = "KwTF60t5OSIHXKsmBee4ZaVYXZykzqMXcQQYN4PcmKZi0j60Um"
        # gridUrl: gridUrl can be found at automation dashboard
        gridUrl = "hub.lambdatest.com/wd/hub"

        # URL: https://{username}:{accessToken}@hub.lambdatest.com/wd/hub
        url = "https://" + username + ":" + accessToken + "@" + gridUrl
        return url



# ----------------------------------------------------------------------
# Robot Parallel runner in docker containers
#
# Assumptions:
# - Executing on a linux vm with all pre-needed software installed
# - Docker container called "robotfwk" already pre-built
# Command line parameters:
# - executionfile : Path to a file containing what to execute
# - suitebase : Path to parent directory containing all suites
# - outputbase : Path to root directory of all outputs
# - maxsuites : Number of max suites to execute in parallel
# ----------------------------------------------------------------------

import argparse
import os.path
import sys
from accelarobotkeywords.runner.parallelrunner import ParallelRunner

def is_true(value) -> bool:
    return value is not None and str(value).strip().upper() == "TRUE"

def main():
    parser = argparse.ArgumentParser(description='Required command line parameters')
    parser.add_argument("-s", "--executionfile", type=str, required=True, help="Path of json file of what to execute")
    parser.add_argument("-b", "--suitebase", type=str, required=True, help="Base path for all suites")
    parser.add_argument("-o", "--outputbase", type=str, required=True, help="Path to outputs root directory")
    parser.add_argument("-m", "--maxsuites", type=int, required=True, help="Max suites to execute at same time")
    parser.add_argument("-p", "--tags", type=str, required=True, help="Robot tags to filter tests to be executed")
    parser.add_argument("-r", "--reporttitle", type=str, required=True, help="Custom Report Title")
    parser.add_argument("-d", "--devmode", type=int, required=False, help="Optional - Is this a dev run")
    parser.add_argument("-t", "--testsuite", type=str, required=False, help="Optional - The single testsuite name")
    parser.add_argument("-l", "--testcases", type=str, required=False, help="Optional - The list of test cases")
    parser.add_argument("-a", "--agency", type=str, required=True, help="Optional - The agency name")
    parser.add_argument("-v", "--civicplatformversion", type=str, required=False, help="Optional - civic platform version")
    parser.add_argument("-e", "--testrailresultenabled", type=str, required=False, help="Optional - test rail result enabled")
    parser.add_argument("-n", "--reruntests", type=int, required=False, help="Optional - rerun test as needed")
    parser.add_argument("-tp", "--testrailprojectid", type=int, required=False, help="Optional - test rail project id")
    parser.add_argument("-ts", "--testrailsuiteid", type=int, required=False, help="Optional - test rail suite case id")
    parser.add_argument("-c", "--confluencefile", type=str, required=False,
                        help="Optional - confluence attachment filename")
    parser.add_argument("-bl", "--browserlocation", type=str, required=False,
                        help="Optional - determine browser location")
    parser.add_argument("-dc", "--desiredcapabilities", type=str, required=False,
                        help="Optional - determine browser desired capabilities")
    parser.add_argument("-z", "--zapenabled", type=str, required=False,
                        help="Optional - determine if zap is enabled")
    parser.add_argument("-ph", "--proxyhost", type=str, required=False,
                        help="Optional - specified proxy host")
    parser.add_argument("-pp", "--proxyport", type=str, required=False,
                        help="Optional - specified proxy port")
    parser.add_argument("-ppa", "--proxypath", type=str, required=False,
                        help="Optional - specified proxy path")
    parser.add_argument("-nl", "--networklogger", type=str, required=False,
                        help="Optional - specified network logger")
    
    args = parser.parse_args()
    print('Execution file = %s' % args.executionfile)
    print('Suites base directory = %s' % args.suitebase)
    print('Output root directory: %s' % args.outputbase)
    print('Max parallel suites: %s' % args.maxsuites)
    if args.tags:
        print('Tags: %s' % args.tags)
    else:
        print('Tags: Robot tag list required')
        sys.exit(1)
    if args.reporttitle:
        print('Custom Report Title: %s' % args.reporttitle)
    else:
        print('Custom Report Title: report title is required')
        sys.exit(1)
    if args.devmode:
        print(f"Dev mode: {args.devmode}")
        if not args.testsuite:
            print("Dev mode: missing test suite name required")
            sys.exit(1)
        if not args.testcases:
            print("Dev mode: missing test case(s) list required")
            sys.exit(1)
        print(f"Test suite: {args.testsuite}")
        print(f"Test Case(s): {args.testcases}")
        print(f"Agency: {args.agency}")
    else:
        args.devmode = 0
        args.testsuite = False
        args.testcases = False
        print("Dev mode: 0 so ignoring any dev mode arguments")

    if not args.reruntests:
        args.reruntests = 0
    print(f"Rerun test: {args.reruntests}")

    tr_enabled = is_true(args.testrailresultenabled)
    args.testrailresultenabled = "TRUE" if tr_enabled else "FALSE"

    if tr_enabled:
        print(f"Test rail result enabled: {args.testrailresultenabled}")
        if not args.civicplatformversion:
            print("Test rail result enabled: missing civic platform version")
            sys.exit(1)

        if args.testrailprojectid and args.testrailsuiteid:
            print(f"Test rail project id {args.testrailprojectid} and Test rail test suite id {args.testrailsuiteid}")
        else:
            v = args.civicplatformversion.lower()
            if ("aca" in v) or ("aca-report" in v) or ("aca-payment" in v) or ("aca-acds" in v):
                args.testrailprojectid = 4
                args.testrailsuiteid = 2379
            elif ("aa" in v) or ("report" in v) or ("payment" in v) or ("acds" in v):
                args.testrailprojectid = 3
                args.testrailsuiteid = 2380
            elif "agis" in v:
                args.testrailprojectid = 101
                args.testrailsuiteid = 2378
            else:
                print("Test rail project/suite not detected from civic platform version; defaulting to AA")
                args.testrailprojectid = 3
                args.testrailsuiteid = 2380

            print(f"Test rail project id {args.testrailprojectid} and Test rail test suite id {args.testrailsuiteid}")
    else:
        args.testrailprojectid = 0
        args.testrailsuiteid = 0
        
    if not args.confluencefile:
        args.confluencefile = False
    if not args.browserlocation:
        args.browserlocation = 'local'
    else:
        print('desired capabilities: %s' % args.desiredcapabilities)
        if args.browserlocation == 'remote':
            if not args.desiredcapabilities:
                args.desiredcapabilities = 'windows.chrome.91'
        else:
                args.browserlocation = 'local'
    print('browser location: %s' % args.browserlocation)
    # check input parameters
    if args.maxsuites < 1:
        print('maxsuites needs to be >1')
        sys.exit(1)
    if not os.path.exists(args.executionfile):
        print('executionfile file not found')
        sys.exit(1)
    if not os.path.isdir(args.suitebase):
        print('suitebase directory not found')
        sys.exit(1)
    if not os.path.isdir(args.outputbase):
        print('outputbase directory not found')
        sys.exit(1)
    if args.zapenabled:
        if args.proxyhost:
            print('proxy host: %s' % args.proxyhost)
        else:
            print('Missing proxy host')
            sys.exit(1)
        if args.proxyport:
            print('proxy port: %s' % args.proxyport)
        else:
            print('Missing proxy port')
            sys.exit(1)
        if args.proxypath:
            print('proxy path: %s' % args.proxypath)
        else:
            print('Missing proxy path')
            sys.exit(1)
    else:
        args.zapenabled = False
        args.proxyport = -1
        args.proxyhost = None
        args.proxypath = None
    if args.networklogger is not None and args.networklogger.upper() == 'TRUE':
        args.networklogger = 'TRUE'
    else:
        args.networklogger = 'FALSE'
    os.environ['NETWORK_LOGGER_ENABLED'] = args.networklogger
    print(f"Network logger enabled: {args.networklogger}")

    # execute
    try:
        runner = ParallelRunner(execution_file=args.executionfile, suite_base_path=args.suitebase,
                                output_base_path=args.outputbase, max_suites=args.maxsuites, robot_tags=args.tags, report_title=args.reporttitle, dev_mode=args.devmode,
                                test_suite_name=args.testsuite, test_case_list=args.testcases, agency=args.agency, civic_platform_version=args.civicplatformversion,
                                test_rail_result_enabled=args.testrailresultenabled, reruntests=args.reruntests, testrail_project_id=args.testrailprojectid,
                                testrail_suite_id=args.testrailsuiteid, confluence_attachment_name=args.confluencefile, browser_location=args.browserlocation,
                                desired_capabilities=args.desiredcapabilities, zap_enabled=args.zapenabled, proxy_host=args.proxyhost, proxy_port=args.proxyport, 
                                proxy_path=args.proxypath, network_logger_enabled=args.networklogger)
        runner.execute_suites()
    except Exception as e:
        print('Error - %s' % str(e))
        sys.exit(1)


if __name__ == "__main__":
    main()



# -----------------------------------------------------------------------------
# Class to load settings from the seleniumsettings.json file
# ----------------------------------------------------------------------------

import json
import os
from robot.api import logger


class SeleniumSettings:
    """
    Class that loads and exposes the selenium settings from the seleniumsettings.json file
    """
    __json_data = None
    __implicit_wait = None
    __explicit_wait = None
    __script_wait = None
    __page_load_wait = None
    __current_browser_name = None
    __smaller_wait = None
    __small_wait = None
    __medium_wait = None
    __long_wait = None
    __extra_long_wait = None

    @staticmethod
    def __get_json_value(key):
        if SeleniumSettings.__json_data is None:
            filename = os.path.abspath(os.path.join(os.path.dirname(__file__), 'seleniumsettings.json'))
            with open(filename, 'r') as json_file:
                raw_data = json_file.read()
                SeleniumSettings.__json_data = json.loads(raw_data)
                logger.info("Selenium settings json file loaded")
        return SeleniumSettings.__json_data[key]

    @staticmethod
    def get_implicit_wait():
        if SeleniumSettings.__implicit_wait is None:
            SeleniumSettings.__implicit_wait = SeleniumSettings.__get_json_value('implicitwait')
        return SeleniumSettings.__implicit_wait

    @staticmethod
    def get_explicit_wait():
        if SeleniumSettings.__explicit_wait is None:
            SeleniumSettings.__explicit_wait = SeleniumSettings.__get_json_value('explicitwait')
        return SeleniumSettings.__explicit_wait

    @staticmethod
    def get_smaller_wait():
        if SeleniumSettings.__smaller_wait is None:
            result = SeleniumSettings.__get_json_value('smallerwait')
            SeleniumSettings.__smaller_wait = int(result)
        logger.info("smallerwait = %s" % str(SeleniumSettings.__smaller_wait))
        return SeleniumSettings.__smaller_wait

    @staticmethod
    def get_small_wait():
        if SeleniumSettings.__small_wait is None:
            result = SeleniumSettings.__get_json_value('smallwait')
            SeleniumSettings.__small_wait = int(result)
        logger.info("smallwait = %s" % str(SeleniumSettings.__small_wait))
        return SeleniumSettings.__small_wait

    @staticmethod
    def get_medium_wait():
        if SeleniumSettings.__medium_wait is None:
            result = SeleniumSettings.__get_json_value('mediumwait')
            SeleniumSettings.__medium_wait = int(result)
        logger.info("mediumwait = %s" % str(SeleniumSettings.__medium_wait))
        return SeleniumSettings.__medium_wait

    @staticmethod
    def get_long_wait():
        if SeleniumSettings.__long_wait is None:
            result = SeleniumSettings.__get_json_value('longwait')
            SeleniumSettings.__long_wait = int(result)
        logger.info("longwait = %s" % str(SeleniumSettings.__long_wait))
        return SeleniumSettings.__long_wait
    
    @staticmethod
    def get_extra_long_wait():
        if SeleniumSettings.__extra_long_wait is None:
            result = SeleniumSettings.__get_json_value('extralongwait')
            SeleniumSettings.__extra_long_wait = int(result)
        logger.debug("extralongwait = %s" % str(SeleniumSettings.__extra_long_wait))
        return SeleniumSettings.__extra_long_wait

    @staticmethod
    def get_script_wait():
        if SeleniumSettings.__script_wait is None:
            result = SeleniumSettings.__get_json_value('scriptwait')
            SeleniumSettings.__script_wait = int(result)
        logger.info("scriptwait = %s" % str(SeleniumSettings.__script_wait))
        return SeleniumSettings.__script_wait

    @staticmethod
    def get_page_load_wait():
        if SeleniumSettings.__page_load_wait is None:
            result = SeleniumSettings.__get_json_value('pageloadwait')
            SeleniumSettings.__page_load_wait = int(result)
        logger.info("pageloadwait = %s" % str(SeleniumSettings.__page_load_wait))
        return SeleniumSettings.__page_load_wait

    @staticmethod
    def set_implicit_wait(seconds):
        SeleniumSettings.__implicit_wait = seconds

    @staticmethod
    def set_explicit_wait(seconds):
        SeleniumSettings.__explicit_wait = seconds

    @staticmethod
    def set_smaller_wait(seconds):
        SeleniumSettings.__smaller_wait = int(seconds)
        logger.info("Smaller wait set to %s seconds" % str(seconds))

    @staticmethod
    def set_small_wait(seconds):
        SeleniumSettings.__small_wait = int(seconds)
        logger.info("Small wait set to %s seconds" % str(seconds))

    @staticmethod
    def set_medium_wait(seconds):
        SeleniumSettings.__medium_wait = int(seconds)
        logger.info("Medium wait set to %s seconds" % str(seconds))

    @staticmethod
    def set_long_wait(seconds):
        SeleniumSettings.__long_wait = int(seconds)
        logger.info("Long wait set to %s seconds" % str(seconds))

    @staticmethod
    def set_extra_long_wait(seconds):
        SeleniumSettings.__extra_long_wait = int(seconds)
        logger.info("Extra long wait set to %s seconds" % str(seconds))

    @staticmethod
    def set_script_wait(seconds):
        SeleniumSettings.__script_wait = int(seconds)
        logger.info('Script wait set to %s seconds' % str(seconds))

    @staticmethod
    def set_page_load_wait(seconds):
        SeleniumSettings.__page_load_wait = int(seconds)
        logger.info('Page load wait set to %s seconds' % str(seconds))

    @staticmethod
    def get_browser_name():
        browsername = os.environ.get('BROWSER_NAME')
        if browsername == None:
            SeleniumSettings.__current_browser_name = 'chrome'
            return SeleniumSettings.__current_browser_name
        browsername = browsername.lower()
        if browsername == 'firefox':
            SeleniumSettings.__current_browser_name = 'firefox'
        elif browsername == 'edge':
            SeleniumSettings.__current_browser_name = 'edge'
        else:
            SeleniumSettings.__current_browser_name = 'chrome'
        logger.debug('BROWSER_NAME is ' + str(SeleniumSettings.__current_browser_name))
        return SeleniumSettings.__current_browser_name



# ---------------------------------------------------------------
# Framework Keywords
#
# This file contains keywords for control of the framework
# ---------------------------------------------------------------

from robot.api import logger
from accelarobotkeywords.config.seleniumsettings import SeleniumSettings
from accelarobotkeywords.config.displaysettings import DisplaySettings
from accelarobotkeywords.framework.framework import Framework


def setup_framework():
    """Main entry point method for the framework"""
    Framework.setup_output_directories()
    Framework.setup_display()


def close_framework():
    """Closes down the test framework"""
    Framework.close_display()


def set_selenium_implicit_wait(seconds):
    """Sets the selenium implicit wait"""
    SeleniumSettings.set_implicit_wait(float(seconds))
    logger.info("Set implicit wait to %s seconds" % str(seconds))


def set_selenium_explicit_wait(seconds):
    """Sets the selenium explicit wait"""
    SeleniumSettings.set_explicit_wait(float(seconds))
    logger.info("Set explicit wait to %s seconds" % str(seconds))


def set_selenium_script_wait(seconds):
    """Sets the selenium javascript wait"""
    SeleniumSettings.set_script_wait(float(seconds))
    logger.info("Set script wait to %s seconds" % str(seconds))


def set_page_load_wait(seconds):
    """Sets the wait time for a page load"""
    SeleniumSettings.set_page_load_wait(float(seconds))
    logger.info("Set page load wait to %s seconds" % str(seconds))


def set_selenium_small_wait(seconds):
    """Sets the number of seconds for the small wait keyword"""
    SeleniumSettings.set_small_wait(seconds)
    logger.info("Small wait set to %s seconds" % seconds)


def set_selenium_medium_wait(seconds):
    """Sets the number of seconds for the medium wait keyword"""
    SeleniumSettings.set_medium_wait(seconds)
    logger.info("Medium wait set to %s seconds" % seconds)


def set_selenium_long_wait(seconds):
    """Sets the number of seconds for the long wait keyword"""
    SeleniumSettings.set_long_wait(seconds)
    logger.info("Long wait set to %s seconds" % seconds)


def set_display_width(width):
    """Sets the width of the displayed browser. Note this needs to be set before the browser is opened"""
    DisplaySettings.set_width(width)
    logger.info("Display width set to: %s" % str(width))


def set_display_height(height):
    """Sets the height of the displayed browser. Note this needs to be set before the browser is opened"""
    DisplaySettings.set_height(height)
    logger.info("Display height set to: %s" % str(height))


def set_display_depth(depth):
    """Sets the color depth of the created display, this keyword is only applicable to docker/linux execution"""
    DisplaySettings.set_depth(depth)
    logger.info("Display depth set to: %s" % str(depth))



# --------------------------------------------------------------------------------------
# Class to generate a virtual display
# --------------------------------------------------------------------------------------

from pyvirtualdisplay import Display
from accelarobotkeywords.config.displaysettings import DisplaySettings
import pyscreenshot as ImageGrab
from robot.api import logger


class VirtualDisplay:
    _display = None

    @staticmethod
    def create_display():
        """Creates a virtual display"""
        width = DisplaySettings.get_width()
        height = DisplaySettings.get_height()
        depth = DisplaySettings.get_depth()
        logger.info("Creating virtual display %s,%s,%s" % (width, height, depth))
        display = Display(backend='xvfb', use_xauth=False, visible=False, color_depth=depth, size=(width, height))
        display.start()
        logger.info("Virtual display created")
        VirtualDisplay._display = display

    @staticmethod
    def stop_display():
        """Stops the virtual display"""
        if VirtualDisplay._display is not None:
            VirtualDisplay._display.stop()
            logger.info("Virtual display was stopped")
        else:
            logger.info("No virtual display to stop")

    @staticmethod
    def capture_screenshot(path):
        """Captures a screenshot to specified path"""
        img = ImageGrab.grab()
        img.save(path)
        logger.info("Screenshot captured to file: %s" % path)





# ---------------------------------------------------------------------------------
# Checkbox Wait Conditions
#
# These are a collection of classes for WebDriverWait specific to Check boxes
#
# --------------------------------------------------------------------------------

from accelarobotkeywords.config.seleniumsettings import SeleniumSettings


class CheckBoxCheck(object):
    """Condition to check/select a check box"""

    def __init__(self, locator):
        self.locator_type = locator.locator_type
        self.locator_value = locator.locator_value

    def __call__(self, driver):
        try:
            driver.implicitly_wait(1)
            element = driver.find_element(self.locator_type, self.locator_value)
            if not element.is_selected():
                element.click()
            return element.is_selected()
        except Exception:
            return False
        finally:
            driver.implicitly_wait(SeleniumSettings.get_implicit_wait())


class CheckBoxUnCheck(object):
    """Condition to uncheck a check box"""

    def __init__(self, locator):
        self.locator_type = locator.locator_type
        self.locator_value = locator.locator_value

    def __call__(self, driver):
        try:
            driver.implicitly_wait(1)
            element = driver.find_element(self.locator_type, self.locator_value)
            if element.is_selected():
                element.click()
            return not element.is_selected()
        except Exception:
            return False
        finally:
            driver.implicitly_wait(SeleniumSettings.get_implicit_wait())


class CheckBoxShouldBeChecked(object):
    """Condition to verify checkbox is checked"""

    def __init__(self, locator):
        self.locator_type = locator.locator_type
        self.locator_value = locator.locator_value

    def __call__(self, driver):
        try:
            driver.implicitly_wait(1)
            element = driver.find_element(self.locator_type, self.locator_value)
            return element.is_selected()
        except Exception:
            return False
        finally:
            driver.implicitly_wait(SeleniumSettings.get_implicit_wait())


class CheckBoxShouldBeUnChecked(object):
    """Condition to verify checkbox is unchecked"""

    def __init__(self, locator):
        self.locator_type = locator.locator_type
        self.locator_value = locator.locator_value

    def __call__(self, driver):
        try:
            driver.implicitly_wait(1)
            element = driver.find_element(self.locator_type, self.locator_value)
            return not element.is_selected()
        except Exception:
            return False
        finally:
            driver.implicitly_wait(SeleniumSettings.get_implicit_wait())



"""
Network Tab Logging Library for Robot Framework Tests

This library provides functionality to capture and print network requests
that occur during test execution using the existing WebDriver instance.
"""

import json
import base64
import html
from typing import Optional, List, Dict, Any
from robot.api import logger as robotlogger
from accelarobotkeywords import SeleniumKeywords


class NetworkLogger:

    ROBOT_LIBRARY_SCOPE = 'GLOBAL'

    def __init__(self):
        self.network_logs = []

    @staticmethod
    def _get_driver_instance():
        driver = SeleniumKeywords._driver
        if driver is None:
            raise Exception("No active WebDriver instance found.")
        try:
            network_logs = driver.get_log('performance')
            robotlogger.info("Retrieved %d log entries" % len(network_logs))
        except Exception as e:
            robotlogger.warn("Performance logging not available: %s. Returning empty list: " % str(e))
            network_logs = []
        
        return driver, network_logs

    @staticmethod
    def _parse_log_entry(log_entry: Dict) -> Optional[Dict[str, Any]]:
        try:
            outer_message = json.loads(log_entry['message'])
            return outer_message['message']
        except (json.JSONDecodeError, KeyError) as e:
            robotlogger.debug("Failed to parse outer_message log entry: %s" % str(e))
            return None

    def _build_requests_map(self, raw_logs: List[Dict]) -> Dict[str, Dict]:
        requests_map = {}

        for log_entry in raw_logs:
            message = self._parse_log_entry(log_entry)
            if not message:
                continue

            method = message.get('method', '')
            params = message.get('params', {})

            if method == 'Network.requestWillBeSent':
                self._handle_request(requests_map, params)
            elif method == 'Network.responseReceived':
                self._handle_response(requests_map, params)

        return requests_map

    @staticmethod
    def _handle_request(requests_map: Dict, params: Dict):
        """Handle Network.requestWillBeSent event."""
        request_id = params.get('requestId', '')
        request = params.get('request', {})
        post_data = request.get('postData', '')
        if post_data and '&' in post_data:
            try:
                post_data = html.unescape(post_data)
            except Exception as e:
                robotlogger.debug("Could not unescape post_data: %s" % str(e))
        
        requests_map[request_id] = {
            'url': request.get('url', ''),
            'method': request.get('method', ''),
            'post_data': post_data,
            'headers': request.get('headers', {}),
            'timestamp': params.get('timestamp', 0),
            'response_received': False
        }

    @staticmethod
    def _handle_response(requests_map: Dict, params: Dict):
        """Handle Network.responseReceived event."""
        request_id = params.get('requestId', '')
        response = params.get('response', {})

        if request_id in requests_map:
            requests_map[request_id].update({
                'response_url': response.get('url', ''),
                'status': response.get('status', 'N/A'),
                'status_text': response.get('statusText', 'N/A'),
                'mime_type': response.get('mimeType', 'N/A'),
                'response_headers': response.get('headers', {}),
                'response_received': True,
                'request_id': request_id
            })

    @staticmethod
    def _matches_filters(req_data: Dict, url_substring: Optional[str],
                        payload_substring: Optional[str]) -> bool:
        if not req_data.get('response_received', False):
            return False

        if url_substring and url_substring not in req_data.get('url', ''):
            return False

        if payload_substring:
            post_data = req_data.get('post_data', '')
            if payload_substring not in post_data:
                return False

        return True

    @staticmethod
    def _get_response_body(driver, request_id: str) -> str:
        BINARY_BODY = "[Binary content - not displayable as text]"
        try:
            result = driver.execute_cdp_cmd(
                "Network.getResponseBody",
                {"requestId": request_id}
            )
            body = result.get("body", "")

            if result.get("base64Encoded", False):
                raw = base64.b64decode(body)

                try:
                    return raw.decode("utf-8")
                except UnicodeDecodeError:
                    try:
                        decoded = raw.decode("latin-1")
                        robotlogger.debug(
                            "Response body for request %s decoded using latin-1" % request_id
                        )
                        return decoded
                    except Exception:
                        return BINARY_BODY
            return body

        except Exception as e:
            return BINARY_BODY

    @staticmethod
    def _format_json(data: str) -> str:
        try:
            parsed = json.loads(data)
            return json.dumps(parsed, indent=2)
        except (json.JSONDecodeError, ValueError):
            return data


    def print_network_requests(self, url_substring: Optional[str] = None,
                               payload_substring: Optional[str] = None) -> List[Dict]:
        driver, raw_logs = self._get_driver_instance()
        self.network_logs = raw_logs

        if not raw_logs:
            robotlogger.info("No network requests captured")
            return []

        robotlogger.info("Processing %d captured network requests" % len(raw_logs))

        requests_map = self._build_requests_map(raw_logs)

        matched_requests = []

        for request_id, req_data in requests_map.items():
            if not self._matches_filters(req_data, url_substring, payload_substring):
                continue

            response_body = self._get_response_body(driver, request_id)
            req_data['response_body'] = response_body

            matched_requests.append(req_data.copy())

        robotlogger.info("Total matches found: %d" % len(matched_requests))


        return matched_requests



import os
from datetime import datetime
from robot import rebot
from accelarobotkeywords.framework.utils.jenkins import Jenkins
from accelarobotkeywords.runner import testrailintegration as testrail
from atlassian import Confluence
import shutil
import docker
import json
import random
import string
import requests
import re
import time

test_dirs = [
    'aa/tests',
    'aca/tests',
    'agis/tests',
    'cam/ui/tests',
    'services/src/tests'
]

class ParallelRunner:
    IMAGE_NAME = 'robotfwk'
    ENTRY_SCRIPT = './execute-suite.sh'
    ROBOT_XML = 'SUITE_output.xml'
    JUNIT_XML = 'SUITE_junit.xml'
    ANNOTATIONS_JSON = "SUITE_annotations.json"
    HTML_REPORT = "allsuites_report.html"
    HTML_LOG = "allsuites_log.html"
    XML_REPORT = "allsuites_junit.xml"

    def __init__(self, execution_file, suite_base_path, output_base_path, max_suites, robot_tags, report_title,
                 dev_mode, test_suite_name, test_case_list, agency, civic_platform_version, test_rail_result_enabled,
                 reruntests, testrail_project_id, testrail_suite_id, confluence_attachment_name, browser_location,
                 desired_capabilities, zap_enabled, proxy_host, proxy_port, proxy_path, network_logger_enabled):
        self.execution_file_path = execution_file
        self.suite_base_path = suite_base_path
        self.output_base_path = output_base_path
        self.max_suites = max_suites
        self.robot_tags = robot_tags
        self.report_title = report_title
        self.dev_mode = dev_mode
        self.test_suite_name = test_suite_name
        self.test_case_list = test_case_list
        self.agency = agency
        self.civic_platform_version = civic_platform_version
        self.test_rail_result_enabled = test_rail_result_enabled
        self.jenkins_build_number = ''
        self.execution_file_json = None
        self.output_path = None
        self.suite_batches = None
        self.execution_name = None
        self.environment_file = None
        self.reruntests = reruntests
        self.testrail_project_id = testrail_project_id
        self.testrail_suite_id = testrail_suite_id
        self.confluence_attachment_name = confluence_attachment_name
        self.browser_location = browser_location
        self.desired_capabilities = desired_capabilities
        self.network_logger_enabled = network_logger_enabled
        self.zap_enabled = zap_enabled
        self.proxy_host = proxy_host
        self.proxy_port = proxy_port
        self.proxy_path = proxy_path

    def execute_suites(self):
        if self.dev_mode > 0:
            if self.test_suite_name.upper() == 'FULL_CICD_SUITE.ROBOT':
                try:
                    print('Executing test case(s) from : %s' % self.test_suite_name)
                    self.__setup_output_directory()
                    cicd_suite = self.__build_dynamic_suite_file()
                    self.execution_file_json = cicd_suite
                    self.__divide_into_batches()
                    batch_count = 1
                    if self.confluence_attachment_name:
                        tr_report = testrail.TestrailIntegration()
                        tr_report.add_testrail_run(self.output_path, self.civic_platform_version, self.jenkins_build_number,
                                       self.test_rail_result_enabled, self.testrail_project_id, self.testrail_suite_id)
                    for batch in self.suite_batches:
                        try:
                            print('Execute batch %s set of containers' % str(batch_count))
                            self.__execute_batch_of_containers(batch)
                            print('Batch %s completed' % str(batch_count))
                            batch_count += 1
                        except Exception:
                            pass
                    self.__merge_container_reports()
                    self.__copy_files_to_jenkins_workspace()
                    print('Execution completed')
                except Exception as e:
                    print('Error executing full CICD suite : %s' % str(e))
                    raise
            else:
                try:
                    print('Executing test case(s) from : %s' % self.test_suite_name)
                    self.__setup_output_directory()
                    single_suite = [
                        {"suitefile": self.test_suite_name, "agency": self.agency, "testcases": self.test_case_list}]
                    self.execution_file_json = {"suites": [
                        {"suitefile": self.test_suite_name, "agency": self.agency, "testcases": self.test_case_list}]}
                    tr_report = testrail.TestrailIntegration()
                    tr_report.add_testrail_run(self.output_path, self.civic_platform_version, self.jenkins_build_number,
                                       self.test_rail_result_enabled, self.testrail_project_id, self.testrail_suite_id)
                    try:
                        print('Execute Suite Test(s) in a container')
                        self.__execute_batch_of_containers(single_suite)
                        print('Single Suite Test(s) is completed')
                    except Exception:
                        pass
                    if  self.reruntests > -1:
                        self.__merge_container_reports()
                        self.__copy_files_to_jenkins_workspace()
                    print('Execution completed')
                except Exception as e:
                    print('Error executing single suite : %s' % str(e))
                    raise
        else:
            print('Executing parallel suites from file: %s' % self.execution_file_path)
            try:
                self.__load_execution_file()
                self.__setup_output_directory()
                self.__divide_into_batches()
                batch_count = 1
                tr_report = testrail.TestrailIntegration()
                tr_report.add_testrail_run(self.output_path, self.civic_platform_version, self.jenkins_build_number,
                                       self.test_rail_result_enabled, self.testrail_project_id, self.testrail_suite_id)
                for batch in self.suite_batches:
                    try:
                        print('Execute batch %s set of containers' % str(batch_count))
                        self.__execute_batch_of_containers(batch)
                        print('Batch %s completed' % str(batch_count))
                        batch_count += 1
                    except Exception:
                        pass
                self.__merge_container_reports()
                self.__copy_files_to_jenkins_workspace()
                print('Execution completed')
            except Exception as e:
                print('Error executing parallel suites : %s' % str(e))
                raise

    def __merge_container_reports(self):
        """Merges the robot framework output files from each container"""
        print('Copying container output xml files to top level')
        files_to_merge = []
        for suite in self.execution_file_json['suites']:
            try:
                if 'suitefile' in suite:
                    name = suite['suitefile'].replace('.robot', '')
                else:
                    name = suite['suitedirectory']
                print('Copying xml file for suite: %s' % name)
                output_xml_path = os.path.join(self.output_path, name, ParallelRunner.ROBOT_XML.replace('SUITE', name))
                destination_path = os.path.join(self.output_path, ParallelRunner.ROBOT_XML.replace('SUITE', name))
                shutil.copyfile(src=output_xml_path, dst=destination_path)
                retries = 6
                while retries > 0:
                    if os.path.exists(output_xml_path) and os.path.getsize(destination_path) > 0:
                        files_to_merge.append(destination_path)
                        break
                    print(f"[WAIT] XML not ready for {name}, retrying...")
                    time.sleep(0.5)
                    retries -= 1
                if retries == 0:
                    print(f"[SKIP] XML for suite {name} still empty after retries.")
            except Exception as e:
                print(f"Error copying XML for suite {suite}: {e}")
        print('Merging container output xml into html report')
        try:
            log_path = os.path.join(self.output_path, 'allsuites_log.html')
            report_path = os.path.join(self.output_path, 'allsuites_report.html')
            junit_path = os.path.join(self.output_path, 'allsuites_junit.xml')
            rebot(*files_to_merge, tagstatinclude='tag TestType:*', name=self.report_title, log=log_path,
                  report=report_path, output=junit_path)
        except Exception as e:
            print('Error merging container xml output: %s' % str(e))
            raise

    def __divide_into_batches(self):
        """Divides the list of suites into batches as per max suites"""
        print('Creating batches for parallel execution')
        num_suites = len(self.execution_file_json['suites'])
        full_batches = num_suites // self.max_suites
        print('- Full batches=%s' % full_batches)
        if num_suites % self.max_suites > 0:
            has_partial = True
        else:
            has_partial = False
        print('- Partial batch at end: %s' % has_partial)
        if has_partial:
            total_batches = full_batches + 1
        else:
            total_batches = full_batches
        print('- %s suites will be divided into %s container batches using max suites %s' % (
            num_suites, total_batches, self.max_suites))
        self.suite_batches = []
        # split full batches
        for batch_counter in range(0, full_batches):
            start_index = batch_counter * self.max_suites
            batch = []
            for counter in range(start_index, start_index + self.max_suites):
                batch.append(self.execution_file_json['suites'][counter])
            self.suite_batches.append(batch)
        print('- full batches created', self.suite_batches)
        # add partial batch
        if has_partial:
            start_index = full_batches * self.max_suites
            batch = []
            for counter in range(start_index, num_suites):
                batch.append(self.execution_file_json['suites'][counter])
            self.suite_batches.append(batch)
            print('- partial batch created', self.suite_batches)

    def __load_execution_file(self):
        """Loads the execution file and checks all suites exist, also basic validation of the json file"""
        with open(self.execution_file_path, 'r') as json_file:
            self.execution_file_json = json.load(json_file)
        # check each suite directory or file exists
        for suite in self.execution_file_json['suites']:
            if 'suitefile' in suite:
                suite_file_path = os.path.join(self.suite_base_path, suite['suitefile'])
                if not os.path.isfile(suite_file_path):
                    print('Suite file not found: %s' % suite_file_path)
                    raise Exception('Suite file not found, aborting: %s' % suite_file_path)
            elif 'suitedirectory' in suite:
                suite_dir_path = os.path.join(self.suite_base_path, suite['suitedirectory'])
                if not os.path.isdir(suite_dir_path):
                    print('Suite directory not found, aborting: %s' % suite_dir_path)
                    raise Exception('Suite directory not found, aborting: %s' % suite_dir_path)
            else:
                print('No suitefile or suitedirectory specified for suite')
                raise Exception('No suitefile or suitedirectory specified for suite, aborting')
        # check suite name specified
        if 'name' in self.execution_file_json:
            print('Execution list name: %s' % self.execution_file_json['name'])
            self.execution_name = self.execution_file_json['name'].replace(' ', '_')
        else:
            print('No execution name was specified in json file')
            raise Exception('No execution name specified in json file, aborting')
        # get environment file name
        if 'environmentfile' in self.execution_file_json:
            print('Execution environment file: %s' % self.execution_file_json['environmentfile'])
            self.environment_file = self.execution_file_json['environmentfile']
        else:
            self.environment_file = 'NotSet'
        print('Suite execution file was loaded')

    def __setup_output_directory(self):
        """Sets up the output directory in the output base path"""
        print('Setting up output directory')
        if Jenkins.is_running_from_jenkins():
            # executing from jenkins
            print('- Executing from jenkins')
            job_name = os.getenv('JOB_NAME')
            build_number = os.getenv('BUILD_NUMBER')
            if len(self.output_base_path) <= 15:
                self.output_path = os.path.join(self.output_base_path, job_name, build_number)
            else:   self.output_path = self.output_base_path
            self.jenkins_build_number = build_number
        else:
            print('- Not executing from Jenkins')
            time_stamp = datetime.now().strftime("%d-%m-%Y-%H-%M-%S")
            self.output_path = os.path.join(self.output_base_path, '%s_%s' % (self.execution_name, time_stamp))
        print('- Creating output directory: %s' % self.output_path)
        if len(self.output_base_path) <= 15:
            try:
                os.makedirs(self.output_path)
            except:
                shutil.rmtree(self.output_path)
                os.mkdir(self.output_path)
        print('- Output directory created')

    def __execute_batch_of_containers(self, batch_suites):
        """Executes and waits for completion of specified list of suite names"""
        print('Executing batch of containers', batch_suites)
        try:
            containers = []
            docker_client = docker.from_env()

            # get the extra_hosts
            extra_hosts_entries = self.__read_extra_hosts()

            print('Robot Tags: %s' % self.robot_tags)
            print('Custom Report Title: %s' % self.report_title)
            for suite in batch_suites:
                self.__handle_suspend_request(self)
                # override the application base path if it is in the suite file
                if 'basepath' in suite:
                    base_path = suite['basepath']
                    print('- Overriding the default basepath for this test suite: %s' % base_path)
                else:
                    suite_base_path = self.suite_base_path
                    parts = os.path.normpath(suite_base_path).split(os.sep)
                    try:
                        basepath_list = parts[parts.index('src') + 1:]
                    except ValueError:
                        basepath_list = []
                    contains_apis_or_ui = False
                    for item in basepath_list:
                        for keyword in ['apis', 'ui', 'val', 'src']:
                            if keyword in item:
                                contains_apis_or_ui = True
                                break
                        if contains_apis_or_ui:
                            break
                    if contains_apis_or_ui:
                        path_list = basepath_list[-3:]
                    else:
                        path_list = basepath_list[-2:]
                    base_path = os.path.join(*path_list)
                    print('Default basepath : %s' % base_path)
                suite_network_logger = suite.get('network_logger', self.network_logger_enabled)
                network_logger_env = 'TRUE' if str(suite_network_logger).upper() in ('TRUE', 'YES') else 'FALSE'
                # get the name of the suite
                if 'suitefile' in suite:
                    name = suite['suitefile']
                    # unique name for each container
                    random_number = ''.join([random.choice(string.ascii_letters + string.digits) for n in range(8)])
                    container_name = name + '_' + random_number
                else:
                    name = suite['suitedirectory']

                print('- Starting container: %s for suite: %s' % (container_name, name))

                # get agency for this suite
                if 'agency' in suite:
                    agency = suite['agency']
                else:
                    agency = self.agency

                if 'robottags' in suite:
                    robottags = suite['robottags']
                else:
                    robottags = self.robot_tags

                # run strategy
                if 'reruntests' in suite:
                    reruntests = suite['reruntests']
                else:   reruntests = self.reruntests

                # get testcases for this suite
                if 'testcases' in suite:
                    testcases = suite['testcases']
                else:
                    testcases = 'NotSet'

                # launch the container
                try:
                    container = docker_client.containers.run(
                        image=ParallelRunner.IMAGE_NAME,
                        auto_remove=True,
                        detach=True,
                        shm_size='512m',
                        volumes={self.output_path: {'bind': '/home/seluser/share', 'mode': 'rw'}},
                        environment={"SUITE": name, "SHARE": "/home/seluser/share", "AGENCY": agency,
                                     "TESTCASES": testcases, "BASEPATH": base_path, "TAGS": robottags,
                                     "REPORTTITLE": self.report_title,
                                     "TESTRAILRUNIDENABLE": self.test_rail_result_enabled,
                                     "TESTRAILPROJECTID": self.testrail_project_id,
                                     "TESTRAILSUITEID": self.testrail_suite_id,
                                     "CIVPVERSION": self.civic_platform_version,
                                     "JENKINSBUILDNUMBER": self.jenkins_build_number,
                                     "RERUNTESTS": reruntests,
                                     "BROWSER_LOCATION": self.browser_location,
                                     "BROWSER_CAPABILITIES": self.desired_capabilities,
                                     "ZAP_ENABLED": self.zap_enabled,  "PROXY_HOST": self.proxy_host,
                                     "PROXY_PORT": self.proxy_port, "PROXY_PATH": self.proxy_path,
                                     "NETWORK_LOGGER_ENABLED": network_logger_env},
                        name=container_name,
                        extra_hosts=extra_hosts_entries,
                        entrypoint=ParallelRunner.ENTRY_SCRIPT
                    )
                    print('- Container started for suite : %s' % name)
                    containers.append(container)
                except Exception as e:
                    print('- Error launching container for suite: %s - error: %s' % (name, str(e)))
            for container in containers:
                print('- Waiting for container: %s' % container.name)
                try:
                    container.wait(timeout=18000)
                    print('- Container for suite: %s, has completed' % container.name)
                except Exception as e:
                    print('- Error determining if container for suite %s has completed' % container.name)
                    print('Exception: ' + str(e))
        except Exception as e:
            print('- Error launching containers: %s' % str(e))

    def __copy_files_to_jenkins_workspace(self):
        """Copies the per suite junit and annotation files to jenkins workspace"""
        if Jenkins.is_running_from_jenkins():
            print('Copying files back to jenkins workspace')
            workspace = Jenkins.get_workspace()
            for suite in self.execution_file_json['suites']:
                if 'suitefile' in suite:
                    name = suite['suitefile'].replace('.robot', '')
                else:
                    name = suite['suitedirectory']
                try:
                    output_junit_path = os.path.join(self.output_path, name,
                                                     ParallelRunner.JUNIT_XML.replace('SUITE', name))
                    destination_path = os.path.join(workspace, ParallelRunner.JUNIT_XML.replace('SUITE', name))
                    print('- Copying junit file: %s' % output_junit_path)
                    shutil.copyfile(src=output_junit_path, dst=destination_path)
                    output_annotations_path = os.path.join(self.output_path, name,
                                                           ParallelRunner.ANNOTATIONS_JSON.replace('SUITE', name))
                    destination_path = os.path.join(workspace, ParallelRunner.ANNOTATIONS_JSON.replace('SUITE', name))
                    print('- Copying annotations file: %s' % output_annotations_path)
                    shutil.copyfile(src=output_annotations_path, dst=destination_path)
                except Exception as e:
                    print('- Error copying file:', e)
            print('- All suite file copied')
            print('- Copying html report files')
            try:
                html_log_path = os.path.join(self.output_path, ParallelRunner.HTML_LOG)
                destination_path = os.path.join(workspace, ParallelRunner.HTML_LOG)
                print('- Copying html log file: %s' % html_log_path)
                shutil.copyfile(src=html_log_path, dst=destination_path)
                html_report_path = os.path.join(self.output_path, ParallelRunner.HTML_REPORT)
                destination_path = os.path.join(workspace, ParallelRunner.HTML_REPORT)
                print('- Copying html report file: %s' % html_report_path)
                shutil.copyfile(src=html_report_path, dst=destination_path)
                xml_report_path = os.path.join(self.output_path, ParallelRunner.XML_REPORT)
                destination_path = os.path.join(workspace, ParallelRunner.XML_REPORT)
                print('- Copying xml junit report file: %s' % xml_report_path)
                shutil.copyfile(src=xml_report_path, dst=destination_path)
            except Exception as e:
                print('- Error copying html,xml files:', e)
        else:
            print('No files to copy to Jenkins workspace')

    def __read_extra_hosts(self):
        """Reads extrahosts.txt file to build out container /etc/hosts file custom entries"""
        try:
            suite_base_path = self.suite_base_path
            parts = os.path.normpath(suite_base_path).split(os.sep)
            try:
                basepath_list = parts[parts.index('src') + 1:]
            except ValueError:
                basepath_list = []
            contains_apis_or_ui = False
            for item in basepath_list:
                for keyword in ['apis', 'ui', 'val', 'src']:
                    if keyword in item:
                        contains_apis_or_ui = True
                        break
                if contains_apis_or_ui:
                    break
            if 'internet' in self.suite_base_path:
                full_file_path = self.suite_base_path + '/../../extrahosts.txt'
            elif contains_apis_or_ui:
                full_file_path = self.suite_base_path + '/../../../../extrahosts.txt'
            else:
                full_file_path = self.suite_base_path + '/../../../extrahosts.txt'
                
            if not os.path.exists(full_file_path) or not os.stat(full_file_path).st_size > 0:
                raise Exception('extrahosts.txt file was not found at %s or file is empty' % full_file_path)
            print('Reading extra hosts file: %s' % full_file_path)
            with open(full_file_path, "r", encoding="utf-8", errors="replace") as hosts_file:
                return self.__build_extra_hosts_dict(hosts_file)
        except OSError as err:
            print('Open extra hosts file failed' + str(err))
            raise
        except Exception as e:
            print('Exception parsing extrahosts.txt file : %s' % e)
            raise

    def __is_test_dir(self, root):
        return any(root.endswith(pattern) for pattern in test_dirs)

    @staticmethod
    def __build_extra_hosts_dict(hosts):
        """Populates and returns an extra_hosts dictionary"""
        try:
            extra_hosts_dict = {}
            for host in hosts:
                host = host.strip()
                if not host or host.startswith('#'):
                    continue
                hostname = host.split()[1].strip()
                ipaddress = host.split()[0].strip()
                extra_hosts_dict.update({hostname: ipaddress})
            return extra_hosts_dict
        except Exception as e:
            raise Exception('Error parsing lines of extrahosts.txt file: %s' % e)

    def __build_dynamic_suite_file(self):
        """builds a suite file dynamically across all application directories"""
        if self.confluence_attachment_name:
            print('- Need to exclude some test cases based on this confluence attachment: %s' % self.confluence_attachment_name)
            self.confluence_json_file = self.__get_json_from_attachment(self, self.confluence_attachment_name)
            if not self.confluence_json_file:
                print('- Error: Unable to build json data object from Confluence attachment named: %s' % self.confluence_attachment_name)
                raise SystemExit()
        else:
            print('- No confluence attachment file specified so will run full set of CICD test cases')
        suite_list = []
        try:
            root_dir = './src'
            for root, dirs, files in os.walk(root_dir):
                if self.__is_test_dir(root):
                    for file_name in files:
                        full_file_path = os.path.join(root, file_name)
                        if self.confluence_attachment_name and file_name.endswith('.robot'):
                            test_case_ids_to_search = self.__get_test_case_ids_to_search(self.confluence_json_file)
                            file_path = full_file_path
                            robot_file_content = None
                            with open(file_path, 'r') as robot_file:
                                robot_file_content = robot_file.read()
                            test_case_ids_in_file = self.__get_test_case_ids(robot_file_content)
                            matching_test_case_ids = test_case_ids_in_file.intersection(test_case_ids_to_search)
                            if matching_test_case_ids:
                                base_path = self.__get_base_path(root)
                                formatted_test_case_ids = self.__format_test_case_ids(list(matching_test_case_ids))
                                suite_list.append('{ "basepath": "%s", "suitefile": "%s", "robottags": "-i %s" },' % (
                                base_path, file_name, formatted_test_case_ids))
                                self.confluence_json_file['testRailIds'] = [test_case for test_case in self.confluence_json_file['testRailIds'] if test_case['id'] not in matching_test_case_ids]
                        else:
                            if file_name.endswith('.robot') and self.__contains_tag(full_file_path):
                                list_of_testcases = self.__get_tagged_testcase_names(self, full_file_path)
                                base_path = self.__get_base_path(root)
                                formatted_test_cases = self.__create_formatted_testcase_list(list_of_testcases)
                                suite_list.append('{ "basepath": "%s", "suitefile": "%s", "testcases": "%s" },' % (
                                    base_path, file_name, formatted_test_cases))
            return self.__create_formatted_suite_file(suite_list)
        except Exception as e:
            print('Exception creating suite list object : %s' % e)
            raise

    @staticmethod
    def __contains_tag(file_path):
        """tests that file contains expected tag or returns False"""
        try:
            for line in open(file_path).readlines():
                if '[TAGS]' in line.upper() and 'CICD:YES' in line.upper():
                    return True
            return False
        except Exception as e:
            print('Exception opening file : %s' % e)
            raise

    def __get_tagged_testcase_names(self, file_path):
        """Returns a list of tagged test case names"""
        testcase_list = []
        try:
            file_lines = []
            with open(file_path, "r", encoding="utf-8") as f:
                file_lines = f.readlines()
            for index, line in enumerate(file_lines):
                if '[TAGS]' in line.upper() and 'CICD:YES' in line.upper():
                    if self.confluence_attachment_name:
                        if self.__verify_test_case_not_excluded(self, line):
                            testcase_list.append(file_lines[index - 2].strip())
                    else:
                        testcase_list.append(file_lines[index - 2].strip())
            return testcase_list
        except Exception as e:
            print('Exception handling file : %s' % e)
            raise

    @staticmethod    
    def __get_test_case_ids(robot_file_content):
        pattern = r'TestRail:(\w+)'
        matches = re.findall(pattern, robot_file_content)
        return set(matches)
    
    @staticmethod    
    def __get_test_case_ids_to_search(json_file):
        test_case_ids_to_search = set()
        for test_case in json_file['testRailIds']:
            if test_case['execute'] == 'Y':
                test_case_ids_to_search.add(test_case['id'])
        return test_case_ids_to_search

    @staticmethod
    def __format_test_case_ids(test_case_ids):
        if len(test_case_ids) == 1:
            return f"TestRail:{test_case_ids[0]}"
        elif len(test_case_ids) > 1:
            return "OR".join(f"TestRail:{test_case_id}" for test_case_id in test_case_ids)
        else:
            return ""

    @staticmethod
    def __verify_test_case_not_excluded(self, tagged_line):
        testrailid_list = re.findall("C[0-9]+", tagged_line)
        for testrailid in testrailid_list:
            if self.__check_test_case_enabled(self.confluence_json_file, testrailid):
                return True
        print("- Unable to find the CICD testrailid(s) in the confluence attachment: %s" % testrailid_list)
        return False

    @staticmethod
    def __get_base_path(root_dir):
        """Resolves the base path that the docker container run command needs"""
        parts = os.path.normpath(root_dir).split(os.sep)
        try:
            basepath_list = parts[parts.index('src') + 1:]
            return os.path.join(*basepath_list)
        except Exception as e:
            print('Exception parsing base path : %s' % e)
            raise

    @staticmethod
    def __create_formatted_testcase_list(list_of_testcases):
        """Creates a formatted list of test cases from list of test cases from a single suite"""
        formatted_test_cases = ''
        for index, test_name in enumerate(list_of_testcases):
            formatted_test_cases = formatted_test_cases + test_name + ","
            if index == len(list_of_testcases) - 1:
                formatted_test_cases = formatted_test_cases[:-1]
        return formatted_test_cases

    @staticmethod
    def __create_formatted_suite_file(suite_list):
        """Creates a formatted suite json file in memory that defines this test run"""
        try:
            formatted_suite_file = ''
            formatted_suite_file = formatted_suite_file + '{'
            formatted_suite_file = formatted_suite_file + '"suites": ['
            for index, line in enumerate(suite_list):
                if index == len(suite_list) - 1:
                    line = line[:-1]
                formatted_suite_file = formatted_suite_file + line
            formatted_suite_file = formatted_suite_file + ']'
            formatted_suite_file = formatted_suite_file + '}'
            formatted_suite_file = json.loads(formatted_suite_file)
        except Exception as e:
            print('Exception creating suite file : %s' % e)
            raise
        return formatted_suite_file

    @staticmethod
    def __get_json_from_attachment(self, confluence_file_name):
        try:
            confluence = self.__get_confluence_object()
            attachments = self.__get_confluence_attachments(confluence)
            for attachment in attachments:
                if attachment['title'] == confluence_file_name + '.json':
                    response = requests.get(confluence.url + attachment['_links']['download'],
                                            auth=(confluence.username, confluence.password))
                    if response.status_code == 200:
                        return response.json()
                else:
                    continue
        except requests.exceptions.RequestException as e:
            raise SystemExit(e)

    @staticmethod
    def __get_confluence_object():
        try:
            accela_url = 'https://accelaeng.atlassian.net/'
            user = 'dvu@accela.com'
            key = 'ATATT3xFfGF0N1h3BU7EJXFQWq4-co0bD-C737glZGChT3teGBP21KP73WWPnODSSo_XF7464ZwEOwW5CgqYKAnghibDsH4HHZIN3PgjqgYc8QhNXmJWQP3VnsjoIv8PuJYR2gFefZdkHdcGZsxywk8AghjHEWfMbgl5O0B8iiQndn_gr68kXLE=A5693CCB'
            return Confluence(url=accela_url, username=user,
                              password=key)
        except requests.exceptions.ConnectionError as conn_err:
            raise SystemExit(conn_err)
        except requests.exceptions.RequestException as e:
            raise SystemExit(e)

    @staticmethod
    def __get_confluence_attachments(confluence):
        try:
            attachments_container = confluence.get_attachments_from_content(page_id=2038988999, start=0, limit=500)
            attachments = attachments_container['results']
            return attachments
        except requests.exceptions.RequestException as e:
            raise SystemExit(e)

    @staticmethod
    def __check_test_case_enabled(json_data, testrailid):
        for index, key in enumerate(json_data['testRailIds']):
            if key['id'] == testrailid and key['execute'] == 'Y':
                return True
        return False

    @staticmethod
    def __handle_suspend_request(self):
        """Handles an external request to temporarily suspend execution"""
        polling_period = 60.0 #60 seconds
        maximum_suspend_minutes = 180  #3 hours
        suspend_file_name = "SUSPENDME"
        suspend_file_path = os.path.join(self.output_path, suspend_file_name)
        if os.path.exists(suspend_file_path):
            print("Suspend file exists, entering polling loop for a maximum of %s minutes...." % maximum_suspend_minutes)
            for i in range(maximum_suspend_minutes):
                time.sleep(polling_period)
                if os.path.exists(suspend_file_path):
                    print("*")
                else:
                    print("Suspend file removed, continuing execution...")
                    return
            print("Polled for maximum allowed time - continuing execution...")
            return
        else:
            return



# ---------------------------------------------------------------------------------
# RadioBox Wait Conditions
#
# These are a collection of classes for WebDriverWait specific to radio boxes
#
# --------------------------------------------------------------------------------

from accelarobotkeywords.config.seleniumsettings import SeleniumSettings


class RadioBoxCheck(object):
    """Condition to check/select a radio box"""
    def __init__(self, locator):
        self.locator_type = locator.locator_type
        self.locator_value = locator.locator_value

    def __call__(self, driver):
        try:
            driver.implicitly_wait(1)
            element = driver.find_element(self.locator_type, self.locator_value)
            element.click()
            return element.is_selected()
        except Exception:
            return False
        finally:
            driver.implicitly_wait(SeleniumSettings.get_implicit_wait())



# -----------------------------------------------------------------------------
# Class to load settings from the seleniumsettings.json file
# ----------------------------------------------------------------------------

import json
import os
from robot.api import logger


class DesiredCapabilities:
    """
    Class that loads and exposes the selenium settings from the seleniumsettings.json file
    """
    __json_data = None
    __desired_cap = None

    @staticmethod
    def __get_json_value(key):
        if DesiredCapabilities.__json_data is None:
            filename = os.path.abspath(os.path.join(os.path.dirname(__file__), 'desiredcapabilities.json'))
            with open(filename, 'r') as json_file:
                raw_data = json_file.read()
                DesiredCapabilities.__json_data = json.loads(raw_data)
                logger.info("Selenium settings json file loaded")
        return DesiredCapabilities.__json_data[key]

    @staticmethod
    def get_browser_desired_capabilities(keyvalue):
        
        try:
            DesiredCapabilities.__desired_cap = DesiredCapabilities.__get_json_value(keyvalue)
            print(DesiredCapabilities.__desired_cap)
            return DesiredCapabilities.__desired_cap
        except KeyError as error:
            raise Exception(error)


# ------------------------------------------------------------
# Class for holding an annotation definition for a test case
# ------------------------------------------------------------


class Annotation:

    def __init__(self, class_name, test_name, test_rail_ids, jira_defect_id, jira_story_id, jira_epic_id):
        self.testPackage = class_name.split('.')[-1]
        self.testClass = class_name
        self.testMethod = test_name
        self.TestRailID = test_rail_ids
        self.JIRADefect = jira_defect_id
        self.JIRAStory = jira_story_id
        self.JIRAEpic = jira_epic_id




# -----------------------------------------------------------------------
# List Wait Conditions
#
# These are a collection of classes for WebDriverWait specific to Lists
#
# ------------------------------------------------------------------------

from accelarobotkeywords.config.seleniumsettings import SeleniumSettings


class ListWaitForItemCount(object):
    """
    An expected condition for waiting until list displays specified number of items
    """
    def __init__(self, locator, count):
        self.locator_type = locator.locator_type
        self.locator_value = locator.locator_value
        self.count = count

    def __call__(self, driver):
        try:
            driver.implicitly_wait(1)
            parent_element = driver.find_element(self.locator_type, self.locator_value)
            list_elements = parent_element.find_elements_by_xpath("li[not(@style='display: none;')]")
            return len(list_elements) == self.count
        except Exception:
            return False
        finally:
            driver.implicitly_wait(SeleniumSettings.get_implicit_wait())



# ------------------------------------------------------------------------
# Class to capture screenshots either from browser or from display
# ------------------------------------------------------------------------

from accelarobotkeywords.framework.framework import Framework
from accelarobotkeywords.config.operatingsystems import OperatingSystem
from accelarobotkeywords.runner.virtualdisplay import VirtualDisplay
from robot.api import logger
from os import path


class ScreenShot:

    @staticmethod
    def capture_screenshot(driver, filename):
        """
        Captures a screenshot to specified file
        If on windows/mac will capture a browser screenshot
        If on linux will capture a display screenshot
        """
        screenshots_dir = Framework.get_screenshots_path()
        filename = filename.replace(' ', '')
        if 'png' not in filename:
            filename = filename + '.png'
        file_path = path.join(screenshots_dir, filename)
        logger.info("Attempting to save screen shot to file: %s" % file_path)
        system = OperatingSystem.get_operating_system()
        if system == OperatingSystem.Linux:
            # capture from display
            VirtualDisplay.capture_screenshot(file_path)
        else:
            driver.get_screenshot_as_file(file_path)
        logger.info("Screen shot saved")



# ---------------------------------------------------------
# OperatingSystem
#
# Defines an enum and method to get the current operating system
#
# -----------------------------------------------------------

from enum import Enum
import platform


class OperatingSystem(Enum):
    Windows = 1,
    Linux = 2,
    Mac = 3

    @staticmethod
    def get_operating_system():
        result = platform.system()
        if result == 'Windows':
            return OperatingSystem.Windows
        elif result == 'Linux':
            return OperatingSystem.Linux
        elif result == 'Darwin':
            return OperatingSystem.Mac
        else:
            raise Exception("Unknown operating system: %s" % result)



# ------------------------------------------------------------------------------
# Class that scans a <table> element
# ------------------------------------------------------------------------------

from robot.api import logger as robotlogger
from accelarobotkeywords.framework.selenium.table.tabledetails import TableDetails


class TableScanner:

    @staticmethod
    def get_table_details(driver, locator):
        """
        Scans the table element and completes the TableDetails object
        :param driver: Seleenium webdriver
        :param locator: the locator for the table element
        """
        robotlogger.info("Scanning html table element")
        result = TableDetails()
        try:
            # get the table element
            table_element = driver.find_element(locator.locator_type, locator.locator_value)
            # check element is really a table
            tag_name = table_element.tag_name.lower()
            if tag_name != 'table':
                raise Exception('Specified element is not a table element')
            # scan column names
            column_names = TableScanner.__scan_columns(table_element)
            if len(column_names) == 0:
                robotlogger.warn('No column names found in table')
            for counter in range(0, len(column_names)):
                result.add_column_name(counter, column_names[counter])
            # scan number of rows
            row_count = TableScanner.__scan_rows(table_element)
            result.set_row_count(row_count)
            # finish
            return result
        except Exception as e:
            robotlogger.error("Error scanning html table: %s" % str(e))
            raise

    @staticmethod
    def search_column_for_value(driver, locator, column_index, search_text, start_row, max_rows):
        """
        Searches a column in the table for the specified search text
        :param driver: Selenium webdriver
        :param locator: Locator of the table element
        :param column_index: column index to search
        :param search_text: text to search in specified column
        :param max_rows: Max number of rows ro search on
        :param start_row: Starting row number
        :return: Row number where earch text was found
        """
        robotlogger.debug("Scanning html table, column %s for value %s" % (str(column_index), search_text))
        try:
            # get the table element
            table_element = driver.find_element(locator.locator_type, locator.locator_value)
            # check element is really a table
            tag_name = table_element.tag_name.lower()
            if tag_name != 'table':
                raise Exception('Specified element is not a table element')
            # perform scan
            row_index = TableScanner.__scan_column_for_value(table_element, column_index, search_text, start_row, max_rows)
            robotlogger.debug('Search text found in row: %s' % row_index)
            return row_index
        except Exception as e:
            robotlogger.error('Error scanning table: %s' % str(e))
            raise

    @staticmethod
    def __scan_columns(table_element):
        """
        Scans the table for columns
        :return Array of column names in order found (their index)
        """
        try:
            # scan for child <th> elements and get their text
            column_names = []
            head_elements = table_element.find_elements(by='xpath', value='descendant::th')
            for head_element in head_elements:
                column_name = head_element.text
                column_names.append(column_name)
            robotlogger.debug("Found table column names: %s" % ' '.join(column_names))
            return column_names
        except Exception as e:
            robotlogger.error("Error scanning column names: %s" % str(e))
            raise

    @staticmethod
    def __scan_rows(table_element):
        """
        Scans the number of rows in the table
        :return: Number of rows found
        """
        try:
            row_elements = table_element.find_elements(by='xpath', value='descendant::tr')
            robotlogger.debug("%s rows found in table" % str(len(row_elements)))
            return len(row_elements)
        except Exception as e:
            robotlogger.error("Error scanning number of rows: %s" % str(e))
            raise

    @staticmethod
    def __scan_column_for_value(table_element, column_index, search_text, start_row, max_rows):
        """
        Scans a table looking for text value in specified column
        :param table_element: WebElement for the the <table> element
        :param column_index: column to search value for
        :param search_text: expected text to search for
        :param max_rows: number of rows to scan or max rows in table
        :param start_row: row number to start from
        :return: row number where cell value found
        """
        try:
            for row_counter in range(start_row, max_rows+1):
                xpath = 'descendant::tr[%s]' % row_counter + '/descendant::td[%s]' % column_index
                cell_element = table_element.find_element(by='xpath', value=xpath)
                cell_text = cell_element.text.strip().lower()
                if cell_text == search_text.lower():
                    robotlogger.debug("Cell value %s found at row %s" % (search_text, str(row_counter)))
                    return row_counter
            raise Exception('Cell value not found')
        except Exception as e:
            robotlogger.error("Error scanning column %s for value %s" % (str(column_index), search_text))
            robotlogger.error(str(e))
            raise



# ---------------------------------------------------------------------
# Annotations listener
#
# This file implements a robot framework listener for test rail ids
# It gets TestRailIds from the TAGS of the test case
# Only Tags beggining with 'C' are considered TestRail ids
# It then writes the annotations file to $ROBOTOUT/annotations.json
# -------------------------------------------------------------------

from accelarobotkeywords.framework.listeners.annotation import Annotation
from robot.api import logger
import json
import os

ROBOT_LISTENER_API_VERSION = 3
ANNOTATIONS_FILE = 'annotations.json'
SUITE_ANNOTATIONS = []


def end_test(data, result):
    """Called at the end of a test case"""
    logger.info('Start of end_test listener')
    try:
        # get robot data
        test_name = data.name
        full_suite_name = data.parent.longname
        tags = data.tags._tags
        # process to annotations data
        class_name = full_suite_name
        testrailids = __parse_test_rail_tags(tags)
        jira_defect = __parse_jira_defect_tags(tags)
        jira_story = __parse_jira_story_tags(tags)
        jira_epic = __parse_jira_epic_tags(tags)
        annotation = Annotation(class_name=class_name, test_name=test_name, test_rail_ids=testrailids,
                                jira_defect_id=jira_defect, jira_story_id=jira_story,
                                jira_epic_id=jira_epic)
        SUITE_ANNOTATIONS.append(annotation)
        logger.info('Test case annotation added to suite annotations')
        logger.info(json.dumps(annotation.__dict__))
    except Exception as e:
        logger.error('Unable to process annotations for test case: %s' % data.name)
        logger.error(e)


def end_suite(data, result):
    """Called at the end of a test suite"""
    logger.info('Start of end_suite listener')
    try:
        json_data = '['
        for item in SUITE_ANNOTATIONS:
            json_data = json_data + json.dumps(item.__dict__) + ','
        json_data = json_data.rstrip(',') + ']'
        robot_out = os.environ.get('ROBOTOUT')
        file_path = os.path.join(robot_out, ANNOTATIONS_FILE)
        with open(file_path, 'w') as json_file:
            json_file.write(json_data)
        logger.info('Annotations file created at: %s' % file_path)
    except Exception as e:
        logger.error('Unable to generate annotations file')
        logger.error(e)


def __parse_test_rail_tags(tags):
    """Parses the tags looking for test rail ids"""
    testrailids = []
    for tag in tags:
        try:
            if 'TestRail:' in tag:
                items = tag.split(':')
                testrailids.append(items[1])
        except Exception:
            logger.warn('Tag: %s, not processed as in wrong format' % tag)
    return testrailids


def __parse_jira_epic_tags(tags):
    """Parses the tags looking for jira epic"""
    for tag in tags:
        try:
            if 'JIRAEpic:' in tag:
                items = tag.split(':')
                return items[1]
        except Exception:
            logger.warn('Tag: %s, not processed as in wrong format' % tag)
    return None


def __parse_jira_story_tags(tags):
    """Parses the tags looking for jira story"""
    for tag in tags:
        try:
            if 'JIRAStory:' in tag:
                items = tag.split(':')
                return items[1]
        except Exception:
            logger.warn('Tag: %s, not processed as in wrong format' % tag)
    return None


def __parse_jira_defect_tags(tags):
    """Parses the tags looking for jira defect"""
    for tag in tags:
        try:
            if 'JIRADefect:' in tag:
                items = tag.split(':')
                return items[1]
        except Exception:
            logger.warn('Tag: %s, not processed as in wrong format' % tag)
    return None







# ---------------------------------------------------------------------------------
# Drop Down Wait Conditions
#
# These are a collection of classes for WebDriverWait specific to drop down lists
#
# --------------------------------------------------------------------------------

from accelarobotkeywords.config.seleniumsettings import SeleniumSettings
from selenium.webdriver.support.select import Select


class DropDownSelectItemByTextValue(object):
    """Condition to select an item in a drop down list by visible text"""
    def __init__(self, locator, text):
        self.locator_type = locator.locator_type
        self.locator_value = locator.locator_value
        self.text = text

    def __call__(self, driver):
        try:
            driver.implicitly_wait(1)
            element = driver.find_element(self.locator_type, self.locator_value)
            select = Select(element)
            select.select_by_visible_text(self.text)
            return True
        except Exception:
            return False
        finally:
            driver.implicitly_wait(SeleniumSettings.get_implicit_wait())


class DropDownSelectItemByValue(object):
    """Condition to select an item in a drop down list by value attribute, eg. value="Individual" """
    def __init__(self, locator, value):
        self.locator_type = locator.locator_type
        self.locator_value = locator.locator_value
        self.value = value

    def __call__(self, driver):
        try:
            driver.implicitly_wait(1)
            element = driver.find_element(self.locator_type, self.locator_value)
            select = Select(element)
            select.select_by_value(self.value)
            return True
        except Exception:
            return False
        finally:
            driver.implicitly_wait(SeleniumSettings.get_implicit_wait())


class DropDownSelectItemByIndex(object):
    """Condition to select an item in a drop down list by index position of the option on the drop down list
     First Item is index 0"""
    def __init__(self, locator, index):
        self.locator_type = locator.locator_type
        self.locator_value = locator.locator_value
        self.index = index

    def __call__(self, driver):
        try:
            driver.implicitly_wait(1)
            element = driver.find_element(self.locator_type, self.locator_value)
            select = Select(element)
            select.select_by_visible_text(select.options[int(self.index)].text)
            return True
        except Exception:
            return False
        finally:
            driver.implicitly_wait(SeleniumSettings.get_implicit_wait())


class DropDownGetCurrentlySelectedOptionText(object):
    """Condition to get currently selected option in drop down """

    def __init__(self, locator):
        self.locator_type = locator.locator_type
        self.locator_value = locator.locator_value

    def __call__(self, driver):
        try:
            driver.implicitly_wait(1)
            element = driver.find_element(self.locator_type, self.locator_value)
            select = Select(element)
            return select.first_selected_option.text
        except Exception:
            return False
        finally:
            driver.implicitly_wait(SeleniumSettings.get_implicit_wait())


class DropDownGetCurrentlySelectedOptionValue(object):
    """Condition to get the value of currently selected option in drop down """

    def __init__(self, locator):
        self.locator_type = locator.locator_type
        self.locator_value = locator.locator_value

    def __call__(self, driver):
        try:
            driver.implicitly_wait(1)
            element = driver.find_element(self.locator_type, self.locator_value)
            select = Select(element)
            return select.first_selected_option.get_attribute("value")
        except Exception:
            return False
        finally:
            driver.implicitly_wait(SeleniumSettings.get_implicit_wait())


class DropDownGetCurrentlySelectedOptionIndex(object):
    """Condition to get the index of currently selected option, for first item it will return 0 """

    def __init__(self, locator):
        self.locator_type = locator.locator_type
        self.locator_value = locator.locator_value

    def __call__(self, driver):
        try:
            driver.implicitly_wait(1)
            element = driver.find_element(self.locator_type, self.locator_value)
            select = Select(element)
            selected_options = select.all_selected_options
            options = select.options
            index = 0
            for option in options:
                for selected_option in selected_options:
                    if option.text == selected_option.text:
                        return index
                index += 1
        except Exception:
            return False
        finally:
            driver.implicitly_wait(SeleniumSettings.get_implicit_wait())


class DropDownGetNumberOfOptions(object):
    """Condition to get total number of available options in drop down """

    def __init__(self, locator):
        self.locator_type = locator.locator_type
        self.locator_value = locator.locator_value

    def __call__(self, driver):
        try:
            driver.implicitly_wait(1)
            element = driver.find_element(self.locator_type, self.locator_value)
            select = Select(element)
            options = select.options
            return len(options)
        except Exception:
            return False
        finally:
            driver.implicitly_wait(SeleniumSettings.get_implicit_wait())


class DropDownGetAllOptions(object):
    """Condition to get All of available options in drop down, it will return a list object  """
    def __init__(self, locator):
        self.locator_type = locator.locator_type
        self.locator_value = locator.locator_value

    def __call__(self, driver):
        try:
            available_options = []
            driver.implicitly_wait(1)
            element = driver.find_element(self.locator_type, self.locator_value)
            select = Select(element)
            options = select.options
            for option in options:
                available_options.append(option.text)
            return available_options
        except Exception:
            return False
        finally:
            driver.implicitly_wait(SeleniumSettings.get_implicit_wait())


# -----------------------------------------------------------------------------------------------------
# Framework
#
# This file contains framework related functions
#
# ----------------------------------------------------------------------------------------------------

import os
from robot.api import logger
from accelarobotkeywords.config.operatingsystems import OperatingSystem
from accelarobotkeywords.runner.virtualdisplay import VirtualDisplay
from os import path


class Framework(object):
    __logs_full_path = None
    __screenshots_full_path = None
    __downloads_full_path = None
    __browser_capabilities = None

    @staticmethod
    def setup_output_directories():
        """
        Creates output directories for the framework under sub directory $ROBOTOUT
        The env variable $ROBOTOUT should be set
        Creates sub-directories:
        - screenshots
        - logs
        Sets class level variables with full path to each
        """
        try:
            robot_out = os.environ.get('ROBOTOUT')
            if robot_out is None:
                raise Exception('ROBOTOUT environment variable not set')
            if not path.exists(robot_out):
                raise Exception('ROBOTOUT directory not found: %s' % robot_out)
            logs_full_path = path.join(robot_out, 'logs')
            screenshots_full_path = path.join(robot_out, 'screenshots')
            downloads_full_path = path.join(robot_out, 'downloads')
            try:
                if not os.path.isdir(logs_full_path):
                    os.mkdir(logs_full_path)
                if not os.path.isdir(screenshots_full_path):
                    os.mkdir(screenshots_full_path)
                if not os.path.isdir(downloads_full_path):
                    os.mkdir(downloads_full_path)
            except:
                pass
            Framework.__logs_full_path = logs_full_path
            Framework.__screenshots_full_path = screenshots_full_path
            Framework.__downloads_full_path = downloads_full_path
            logger.info("Created logs directory: %s" % logs_full_path)
            logger.info("Created screenshots directory: %s" % screenshots_full_path)
            logger.info("Created downloads directory: %s" % downloads_full_path)
        except Exception as e:
            logger.error('Unable to create output directories')
            logger.error(str(e))
            raise

    @staticmethod
    def get_logs_path():
        return Framework.__logs_full_path

    @staticmethod
    def get_screenshots_path():
        return Framework.__screenshots_full_path

    @staticmethod
    def get_downloads_path():
        return Framework.__downloads_full_path
    
    @staticmethod
    def get_browser_capabilities():
        return Framework.__browser_capabilities

    @staticmethod
    def setup_display():
        """Sets up any needed virtual display"""
        system = OperatingSystem.get_operating_system()
        if system == OperatingSystem.Linux:
            VirtualDisplay.create_display()
        else:
            logger.info("Not on linux so no virtual display needed")

    @staticmethod
    def close_display():
        """Closes down any virtual display"""
        system = OperatingSystem.get_operating_system()
        if system == OperatingSystem.Linux:
            VirtualDisplay.stop_display()
        else:
            logger.info("Not on linux so no virtual display needed")
            


# ---------------------------------------------------------
# Browser Locations
# -----------------------------------------------------------

from enum import Enum


class BrowserLocations(Enum):
    Local = 1,
    Remote = 2
    
    @staticmethod
    def get_browser_location(browser_location):
        try:
            if browser_location.lower()=='remote':
                return BrowserLocations.Remote
            else:
                return BrowserLocations.Local
        except:
            return BrowserLocations.Local


# -------------------------------------------------------
# Custom locator string class
#
# This class parses a string into a valid locator
# The string format is: <locator-type>:<locator-value>
# With possible locator types as:
# - id
# - xpath
# - link-text
# - partial-link-text
# - name
# - tag-name
# - class-name
# - css-selector
#
# -------------------------------------------------------


class LocatorParseError(Exception):
    """ Exception for parsing a locator string """
    pass


class StringLocator(object):
    __valid_locator_types = ["id", "xpath", "link-text", "partial-link-text", "name", "tag-name", "class-name", "css-selector"]

    def __init__(self, locator_string):
        self.locator_type, self.locator_value = self.__parse_string(locator_string)

    # Parses the string into two parts, locator-type, locator-value
    # Throws a LocatorParseError if cannot parse the string
    def __parse_string(self, locator_string):
        try:
            locator_parts = locator_string.split(':', 1)
            parts_count = len(locator_parts)
            if parts_count != 2:
                raise Exception("Invalid locator, type or value not specified")
            locator_type = locator_parts[0]
            locator_value = locator_parts[1]
            if locator_type not in self.__valid_locator_types:
                raise Exception("Invalid locator type: %s" % locator_type)
            if len(locator_value) <= 1:
                raise Exception("Invalid locator value: %s" % locator_value)
            selenium_locator_type = locator_type.replace('-', ' ')
            return selenium_locator_type, locator_value
        except Exception as e:
            raise LocatorParseError(e)
